#   Copyright ETH 2023 - 2024 Zürich, Scientific IT Services
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
import binascii
import functools
import json
import os
import time
from pathlib import Path
from threading import Lock, Thread
from urllib.parse import urljoin

import requests
from requests.adapters import HTTPAdapter, Retry

DSS_V3 = "/datastore_server/rmi-data-store-server-v3.json"
REQUEST_RETRIES_COUNT = 3
DOWNLOAD_RETRIES_COUNT = 3
FAST_DOWNLOAD_PROTOCOL_VERSION = 2


def make_fileserver_body_params(server_information, **params):
    """create a proper pam of key-values for fileserver request"""
    result = {}
    if server_information.is_version_greater_than(3, 6):
        result = {
            "version": [str(FAST_DOWNLOAD_PROTOCOL_VERSION)]
        }

    for key, value in params.items():
        result[str(key)] = [str(value).replace("'", '"')]
    return result


def comma_separated_items(arr):
    """Create comma-separated string from the list of items"""
    return functools.reduce(lambda a, b: a + ", " + b, arr)


def create_session(download_url_base):
    """Create a session object to handle retries in case of server failure"""
    session = requests.Session()
    retries = Retry(total=REQUEST_RETRIES_COUNT, backoff_factor=1, status_forcelist=[502, 503, 504])
    session.mount(download_url_base, HTTPAdapter(max_retries=retries))
    return session


def post_request(session, full_url, verify_certificates, request, parse_response=True):
    """Perform POST call to server"""
    try:
        if request:
            resp = session.post(full_url, json.dumps(request), verify=verify_certificates)
        else:
            resp = session.post(full_url, verify=verify_certificates)
    except requests.exceptions.SSLError as exc:
        raise requests.exceptions.SSLError(
            "Certificate validation failed. Use o=Openbis(url, verify_certificates=False) if you are using self-signed certificates."
        ) from exc
    except requests.ConnectionError as exc:
        raise requests.ConnectionError(
            "Could not connecto to the openBIS server. Please check your internet connection, the specified hostname and port."
        ) from exc
    if resp.ok:
        if parse_response is True:
            resp = resp.json()
            if "error" in resp:
                print(json.dumps(resp))
                raise ValueError(resp["error"]["message"])
        return resp
    else:
        raise ValueError("general error while performing post request")


def queue_chunks(session, base_url, download_session_id, chunks, verify_certificates, server_information):
    """Queue particular session chunks for download"""
    queue_request = make_fileserver_body_params(server_information,
                                                method='queue',
                                                downloadSessionId=download_session_id,
                                                ranges=comma_separated_items(chunks))
    response = post_request(session, base_url, verify_certificates, queue_request,
                            parse_response=False)
    if response.ok is False:
        raise ValueError(
            "error during queueing for download! Message:" + response["error"]["message"])


def deserialize_chunk(byte_array):
    sequence_number_bytes = 4
    download_item_id_length_bytes = 2
    is_directory_bytes = 1
    file_path_length_bytes = 2
    file_offset_bytes = 8
    payload_length_bytes = 4
    sent_header_checksum_bytes = 8
    sent_payload_checksum_bytes = 8

    result = {
        "invalid": False,
        'invalid_reason': ""
    }

    if len(byte_array) == 0:
        result['invalid'] = True
        result['invalid_reason'] = "HEADER"
        return result

    start, end = 0, sequence_number_bytes
    result['sequence_number'] = int.from_bytes(byte_array[start:end], "big")
    start, end = end, end + download_item_id_length_bytes
    download_item_id_length = int.from_bytes(byte_array[start:end], "big")
    start, end = end, end + is_directory_bytes
    result['is_directory'] = bool.from_bytes(byte_array[start:end], "big")
    start, end = end, end + file_path_length_bytes
    file_path_length = int.from_bytes(byte_array[start:end], "big")
    start, end = end, end + file_offset_bytes
    result['file_offset'] = int.from_bytes(byte_array[start:end], "big")
    start, end = end, end + payload_length_bytes
    result['payload_length'] = int.from_bytes(byte_array[start:end], "big")
    start, end = end, end + download_item_id_length
    result['download_item_id'] = byte_array[start:end].decode("utf8")
    start, end = end, end + file_path_length
    result['file_path'] = byte_array[start:end].decode("utf8")
    calculated_header_checksum = binascii.crc32(byte_array[:end])
    # End of header
    start, end = end, end + sent_header_checksum_bytes
    sent_header_checksum = int.from_bytes(byte_array[start:end], "big")
    if sent_header_checksum != calculated_header_checksum:
        result['invalid'] = True
        result['invalid_reason'] = "HEADER"
        return result
    start, end = end, end + result['payload_length']
    result['payload'] = byte_array[start:end]
    calculated_payload_checksum = binascii.crc32(byte_array[start:end])
    # End of payload
    start, end = end, end + sent_payload_checksum_bytes
    sent_payload_checksum = int.from_bytes(byte_array[start:end], "big")
    if sent_payload_checksum != calculated_payload_checksum:
        result['invalid'] = True
        result['invalid_reason'] = "PAYLOAD"
        return result

    return result


class AtomicChecker:
    """Helper class for keeping watch of chunks to download"""

    def __init__(self, values_to_download: set):
        self._value = 0
        self._max = len(values_to_download)  # limit to not have an infinite download sessions
        self._set = values_to_download
        self._lock = Lock()

    def should_continue(self):
        with self._lock:
            if self._value >= self._max:
                return False
            self._value += 1
            return True

    def repeat_call(self):
        with self._lock:
            self._max += 1

    def break_count(self):
        with self._lock:
            self._max = 0

    def remove_value(self, value):
        with self._lock:
            if value in self._set:
                self._set.remove(value)

    def get_remaining_values(self):
        return self._set


def _get_json(response):
    try:
        return True, response.json()
    except:
        return False, response


class DownloadThread(Thread):
    """Helper class defining single stream download"""

    def __init__(self, session, download_url_base, download_session_id, stream_id,
                 counter: AtomicChecker, verify_certificates, create_default_folders, destination,
                 server_information, number_of_chunks=1):
        Thread.__init__(self)
        self.session = session
        self.download_url = download_url_base
        self.download_session_id = download_session_id
        self.stream_id = stream_id
        self.counter = counter
        self.number_of_chunks = number_of_chunks
        self.create_default_folders = create_default_folders
        self.destination = destination
        self.verify_certificates = verify_certificates
        self.exc = None
        self.server_information = server_information

    def run(self):
        repeated_chunks = {}
        download_params = make_fileserver_body_params(self.server_information,
                                                      method='download',
                                                      downloadSessionId=self.download_session_id,
                                                      numberOfChunks=self.number_of_chunks,
                                                      downloadStreamId=self.stream_id)
        retry_counter = 0
        while self.counter.should_continue():
            try:
                download_response = self.session.post(self.download_url,
                                                      data=json.dumps(download_params), stream=True,
                                                      verify=self.verify_certificates)
                if download_response.ok is True:
                    data = deserialize_chunk(download_response.content)
                    if data['invalid'] is True:
                        is_json, response = _get_json(download_response)
                        if is_json:
                            if 'retriable' in response and response['retriable'] is False:
                                self.counter.break_count()
                                raise ValueError(response["error"])
                        else:
                            if data['invalid_reason'] == "PAYLOAD":
                                sequence_number = data['sequence_number']
                                if repeated_chunks.get(sequence_number, 0) >= DOWNLOAD_RETRIES_COUNT:
                                    self.counter.break_count()
                                    raise ValueError(
                                        "Received incorrect payload multiple times. Aborting.")
                                repeated_chunks[sequence_number] = repeated_chunks.get(sequence_number,
                                                                                       0) + 1
                                queue_chunks(self.session, self.download_url,
                                             self.download_session_id,
                                             [f"{sequence_number}:{sequence_number}"],
                                             self.verify_certificates, self.server_information)
                                self.counter.repeat_call()  # queue additional download chunk run

                        if retry_counter >= REQUEST_RETRIES_COUNT:
                            self.counter.break_count()
                            raise ValueError("Consecutive download calls to the server failed.")

                        # Exponential backoff for the consecutive failures
                        time.sleep(2 ** retry_counter)
                        retry_counter += 1

                    else:
                        retry_counter = 0
                        sequence_number = data['sequence_number']
                        self.save_to_file(data)
                        self.counter.remove_value(sequence_number)
            except Exception as e:
                self.exc = e
        return True

    def save_to_file(self, deserialized_response):
        file_name = deserialized_response['file_path']
        if self.create_default_folders:
            # create original/ or original/DEFAULT subdirectories
            filename_dest = os.path.join(self.destination, file_name)
        else:
            # ignore original/ and original/DEFAULT folders that come from openBIS
            if file_name.startswith("original/"):
                file_name = file_name.replace("original/", "", 1)
            if file_name.startswith("DEFAULT/"):
                file_name = file_name.replace("DEFAULT/", "", 1)
            filename_dest = os.path.join(self.destination, file_name)

        # create the necessary directory structure if they don't exist yet
        os.makedirs(os.path.dirname(filename_dest), exist_ok=True)

        if deserialized_response['is_directory'] is False:
            # create file if it does not exist already
            Path(filename_dest).touch(exist_ok=True)

            file_offset = deserialized_response['file_offset']
            with open(filename_dest, "rb+") as file:
                file.seek(file_offset)
                file.write(deserialized_response['payload'])


class FastDownload:
    """Class for downloading data using FastDownload scheme"""

    def __init__(
            self,
            token,
            download_url,
            perm_id,
            files,
            destination,
            create_default_folders,
            wait_until_finished,
            verify_certificates,
            server_information,
            wished_number_of_streams=4
    ):
        self.dss_facade_url = urljoin(download_url, DSS_V3)
        self.session = create_session(download_url)
        self.token = token
        self.verify_certificates = verify_certificates
        self.perm_id = perm_id
        self.destination = destination
        self.create_default_folders = create_default_folders
        self.wait_until_finished = wait_until_finished
        self.wished_number_of_streams = wished_number_of_streams
        self.server_information = server_information

        if files is None:
            raise ValueError("please provide at least one file")

        if isinstance(files, str):
            files = [files]

        self.files = files

    def download(self):
        """Fast download of files from dataset"""

        if self.token is None:
            raise ValueError("Your session expired, please log in again")

        # Step 1 - request DSS Facade to create fast download session in fileserver

        create_fast_download_response = \
            post_request(self.session, self.dss_facade_url, self.verify_certificates,
                         self._create_fast_download_session_request())['result']
        download_url = create_fast_download_response['downloadUrl']
        user_session_id = create_fast_download_response['fileTransferUserSessionId']

        # Step 2 - Request fileserver to start the download session

        if self.server_information.is_version_greater_than(3, 6):
            download_item_ids = list(map(lambda file: file['filePath'], create_fast_download_response['files']))
        else:
            download_item_ids = comma_separated_items(
                map(lambda file: file['filePath'], create_fast_download_response['files']))

        start_session_params = make_fileserver_body_params(self.server_information,
                                                           method="startDownloadSession",
                                                           userSessionId=user_session_id,
                                                           downloadItemIds=download_item_ids,
                                                           wishedNumberOfStreams=self.wished_number_of_streams)

        start_download_session = post_request(self.session, download_url,
                                              self.verify_certificates,
                                              start_session_params)
        download_session_id = start_download_session['downloadSessionId']


        # Step 3 - Put files into fileserver download queue

        ranges = start_download_session['ranges']
        self._queue_all_files(download_url, download_session_id, ranges)

        # Step 4 & 5 - Download files in chunks and close connection

        session_stream_ids = list(start_download_session['streamIds'])

        exception_list = []
        thread = Thread(target=self._download_step,
                        args=(download_url, download_session_id, session_stream_ids, ranges,
                              exception_list))
        thread.start()

        if self.wait_until_finished is True:
            thread.join()
            if exception_list:
                raise exception_list[0]

        return self.destination

    def _create_fast_download_session_request(self):
        file_ids = list(
            map(lambda file_path: self._make_json_id(file_path), self.files))

        fast_download_session_options = {
            "wishedNumberOfStreams": self.wished_number_of_streams,
            "@type": "dss.dto.datasetfile.fastdownload.FastDownloadSessionOptions",
        }

        return {
            "id": "2",
            "jsonrpc": "2.0",
            "method": "createFastDownloadSession",
            "params": [self.token, file_ids, fast_download_session_options]
        }

    def _make_json_id(self, file_path):
        """Prepare JSON to create session for fileserver for given file in dataset"""
        return {
            "dataSetId": {
                "permId": self.perm_id,
                "@type": "as.dto.dataset.id.DataSetPermId"
            },
            "filePath": self.perm_id + "/" + file_path,
            "@type": "dss.dto.datasetfile.id.DataSetFilePermId"
        }

    def _queue_all_files(self, base_url, download_session_id, ranges):
        """
        queue all chunks for download from fileserver, each file receives different chunk range
        FileA: 0:4
        FileB: 5:6
        """
        chunks = []
        for file, chunks_range in ranges.items():
            chunks += [chunks_range]
        queue_chunks(self.session, base_url, download_session_id, chunks,
                     self.verify_certificates, self.server_information)

    def _download_step(self, download_url, download_session_id, session_stream_ids, ranges,
                       exception_list):
        """
        Perform downloading of chunks in separate threads
        :param download_url: url to use for downloading data
        :param download_session_id: download session id
        :param session_stream_ids: list of available streams
        :param ranges: ranges provided for files
        :return: nothing
        """
        min_chunk = min(map(lambda x: int(x.split(":")[0]), ranges.values()))
        max_chunk = max(map(lambda x: int(x.split(":")[1]), ranges.values()))
        chunks_to_download = set(range(min_chunk, max_chunk + 1))

        counter = 1
        try:
            while True:  # each iteration will create threads for streams
                checker = AtomicChecker(chunks_to_download)
                streams = [
                    DownloadThread(self.session, download_url, download_session_id, stream_id, checker,
                                   self.verify_certificates, self.create_default_folders,
                                   self.destination, self.server_information) for stream_id in session_stream_ids]

                for thread in streams:
                    thread.start()
                for thread in streams:
                    thread.join()

                if chunks_to_download == set():  # if there are no more chunks to download
                    break
                else:
                    if counter >= DOWNLOAD_RETRIES_COUNT:
                        print(f"Reached maximum retry count:{counter}. Aborting.")
                        exception_list += [
                            ValueError(f"Reached maximum retry count:{counter}. Aborting.")]
                        break
                    exceptions = [stream.exc for stream in streams if stream.exc is not None]
                    if exceptions:
                        print(f"Download failed with message: {exceptions[0]}")
                        exception_list += exceptions
                        break
                    counter += 1
                    # queue chunks that failed to download in the previous pass
                    queue_chunks(self.session, download_url, download_session_id,
                                 [f"{x}:{x}" for x in chunks_to_download],
                                 self.verify_certificates, self.server_information)
        finally:
            # Step 5 - Close the session
            finish_download_session_params = make_fileserver_body_params(self.server_information,
                method='finishDownloadSession',
                downloadSessionId=download_session_id)

            self.session.post(download_url,
                              data=json.dumps(finish_download_session_params),
                              verify=self.verify_certificates)



#   Copyright ETH 2018 - 2023 Zürich, Scientific IT Services
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#        http://www.apache.org/licenses/LICENSE-2.0
#   
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
import base64
import os
import pathlib

from pandas import DataFrame
from tabulate import tabulate

from .attachment import Attachment
from .definitions import (
    openbis_definitions,
    fetch_option,
    get_method_for_entity,
    get_type_for_entity,
)
from .utils import (
    format_timestamp,
    nvl,
    extract_person,
)


class AttrHolder:
    """General class for both samples and experiments that hold all common attributes, such as:
    - space
    - project
    - experiment (sample)
    - samples (experiment)
    - dataset
    - tags
    """

    def __init__(self, openbis_obj, entity, type=None):
        self.__dict__["_openbis"] = openbis_obj
        self.__dict__["_entity"] = entity

        if type is not None:
            self.__dict__["_type"] = type.data

        self.__dict__["_defs"] = openbis_definitions(entity)
        # self.__dict__['_allowed_attrs'] = openbis_definitions(entity)['attrs']
        # self.__dict__['_allowed_attrs_new'] = openbis_definitions(entity)['attrs_new']
        # self.__dict__['_allowed_attrs_up'] = openbis_definitions(entity)['attrs_up']
        self.__dict__["_identifier"] = None
        self.__dict__["_is_new"] = True
        self.__dict__["_tags"] = []

    def __call__(self, data):
        """This internal method is invoked when an existing object is loaded.
        Instead of invoking a special method we «call» the object with the data
           self(data)
        which automatically invokes this method.
        Since the data comes from openBIS, we do not have to check it (hence the
        self.__dict__ statements to prevent invoking the __setattr__ method)
        Internally data is stored with an underscore, e.g.
            sample._space = {
                '@type': 'as.dto.space.id.SpacePermId',
                'permId': 'MATERIALS'
            }
        but when fetching the attribute without the underscore, we only return
        the relevant data for the user:
            sample.space   # MATERIALS
        """
        # entity is read from openBIS, so it is not new anymore
        self.__dict__["_is_new"] = False
        if 'fetchOptions' in data:
            self.__dict__["_fetchOptions"] = data['fetchOptions']

        for attr in self._defs["attrs"]:
            if attr in ["code", "permId", "identifier", "type"]:
                self.__dict__["_" + attr] = data.get(attr, None)
                # remove the @id attribute
                if isinstance(self.__dict__["_" + attr], dict):
                    self.__dict__["_" + attr].pop("@id", None)

            elif attr in ["vocabularyCode"]:
                self.__dict__["_" + attr] = data.get("permId", {}).get(attr, None)

            elif attr in ["validationPlugin"]:
                d = data.get(attr, None)
                if d is not None:
                    d = d["permId"]
                self.__dict__["_" + attr] = d

            elif attr in ["space"]:
                d = data.get(attr, None)
                if d is not None:
                    d = d["permId"]
                self.__dict__["_" + attr] = d

            elif attr in ["sample", "experiment", "project", "container"]:
                d = data.get(attr, None)
                if d is not None:
                    d = d["identifier"]
                self.__dict__["_" + attr] = d

            elif attr in ["parents", "children", "samples", "components", "containers"]:
                self.__dict__["_" + attr] = []
                if data[attr] is not None:
                    for item in data[attr]:
                        try:
                            if "identifier" in item:
                                self.__dict__["_" + attr].append(item["identifier"])
                            elif "permId" in item:
                                self.__dict__["_" + attr].append(item["permId"])
                        except Exception:
                            # TODO: under certain circumstances, openBIS only delivers an integer
                            pass

            elif attr in ["tags"]:
                self.add_tags(data[attr])

            elif attr.endswith("Date"):
                self.__dict__["_" + attr] = format_timestamp(data.get(attr))

            elif attr in ["registrator", "modifier", "dataProducer", "owner"]:
                self.__dict__["_" + attr] = extract_person(data.get(attr))

            elif attr in ["sampleType"] and self.entity == 'propertyType' and data.get('dataType', None) == 'SAMPLE':
                if data.get('sampleType', None) is None:
                    self.__dict__["_" + attr] = "(ALL)"
                else:
                    self.__dict__["_" + attr] = data.get(attr, None)

            else:
                self.__dict__["_" + attr] = data.get(attr, None)

    def _new_attrs(self, method_name=None):
        """Returns the Python-equivalent JSON request when a new object is created.
        It is used internally by the save() method of a newly created object.
        """
        attr2ids = openbis_definitions("attr2ids")
        new_obj = get_type_for_entity(self.entity, "create")

        for attr in self._defs["attrs_new"]:
            items = None

            if attr == "type":
                new_obj["typeId"] = self._type["permId"]
                continue

            elif attr == "kind":
                # when creating a new dataset, the attribute «kind» is called «dataSetKind»
                new_obj["dataSetKind"] = self._kind

            elif attr == "attachments":
                attachments = getattr(self, "_new_attachments")
                if attachments is None:
                    continue
                atts_data = [attachment.get_data() for attachment in attachments]
                items = atts_data

            elif attr == "userIds":
                if "_changed_users" not in self.__dict__:
                    continue

                new_obj[attr] = []
                for userId in self.__dict__["_changed_users"]:
                    if self.__dict__["_changed_users"][userId]["action"] == "Add":
                        new_obj[attr].append(
                            {"permId": userId, "@type": "as.dto.person.id.PersonPermId"}
                        )

            else:
                items = getattr(self, "_" + attr)

                key = None
                if attr in attr2ids:
                    # translate parents into parentIds, children into childIds etc.
                    key = attr2ids[attr]
                else:
                    key = attr

                new_obj[key] = items

        # if method_name is not defined: guess the method name for creating a new entity
        if method_name is None:
            method_name = get_method_for_entity(self.entity, "create")
        request = {"method": method_name, "params": [self.openbis.token, [new_obj]]}
        return request

    def _up_attrs(self, method_name=None, permId=None):
        """Returns the Python-equivalent JSON request when a new object is updated.
        It is used internally by the save() method of an object to be updated.
        """
        # defs = openbis_definitions(self._entity)
        attr2ids = openbis_definitions("attr2ids")

        up_obj = get_type_for_entity(self.entity, "update")

        # for some weird reasons, the permId is called differently
        # for every openBIS entity, but only when updating...
        identifier_name = self._defs["identifier"]
        if permId:
            up_obj[identifier_name] = permId
        else:
            up_obj[identifier_name] = self._permId

        # look at all attributes available for that entity
        # that can be updated
        for attr in self._defs["attrs_up"]:
            items = None

            if attr == "attachments":
                # v3 API currently only supports adding attachments
                attachments = self.__dict__.get("_new_attachments", None)
                if attachments is None:
                    continue
                atts_data = [attachment.get_data() for attachment in attachments]

                up_obj["attachments"] = {
                    "actions": [
                        {
                            "items": atts_data,
                            "@type": "as.dto.common.update.ListUpdateActionAdd",
                        }
                    ],
                    "@type": "as.dto.attachment.update.AttachmentListUpdateValue",
                }

            elif attr == "tags":
                items = []
                for tag in self.__dict__["_tags"]:
                    items.append(
                        {"permId": tag["permId"], "@type": "as.dto.tag.id.TagPermId"}
                    )

                up_obj["tagIds"] = {
                    "actions": [
                        {
                            "items": items,
                            "@type": "as.dto.common.update.ListUpdateActionSet",
                        }
                    ],
                    "@type": "as.dto.common.update.IdListUpdateValue",
                }
            elif attr == "metaData":
                # ListUpdateMapValues
                metaData = self.__dict__["_" + attr]
                if metaData:
                    items = [metaData]
                    data_type = "as.dto.common.update.ListUpdateActionSet"
                elif metaData is not None and len(metaData) == 0:
                    # metaData needs to be set to {} in order to remove it.
                    items = ["custom_widget"]
                    data_type = "as.dto.common.update.ListUpdateActionRemove"
                else:
                    items = [{}]
                    data_type = "as.dto.common.update.ListUpdateActionSet"
                up_obj[attr] = {
                    "actions": [{"items": items, "@type": data_type}],
                    "@type": "as.dto.common.update.ListUpdateMapValues",
                }
            elif attr == "userIds":
                actions = []
                if "_changed_users" not in self.__dict__:
                    continue
                for userId in self.__dict__["_changed_users"]:
                    actions.append(
                        {
                            "items": [
                                {
                                    "permId": userId,
                                    "@type": "as.dto.person.id.PersonPermId",
                                }
                            ],
                            "@type": f'as.dto.common.update.ListUpdateAction{self.__dict__["_changed_users"][userId]["action"]}',
                        }
                    )

                up_obj["userIds"] = {
                    "actions": actions,
                    "@type": "as.dto.common.update.IdListUpdateValue",
                }

            elif (
                attr
                in "description label official ordinal autoGeneratedCode subcodeUnique listable showContainer showParents showParentMetadata disallowDeletion validationPlugin pattern patternType script".split()
            ):
                # alway update common fields
                key = attr2ids.get(attr, attr)
                up_obj[key] = {
                    "value": self.__dict__["_" + attr],
                    "isModified": True,
                    "@type": "as.dto.common.update.FieldUpdateValue",
                }

            elif "_" + attr in self.__dict__:
                # handle multivalue attributes (parents, children, tags etc.)
                # we only cover the Set mechanism, which means we always update
                # all items in a list
                if "multi" in self._defs and attr in self._defs["multi"]:
                    items = self.__dict__.get("_" + attr, [])
                    if items is None:
                        # 'None' multivalue attributes should be omitted during update
                        continue
                    up_obj[attr2ids[attr]] = {
                        "actions": [
                            {
                                "items": items,
                                "@type": "as.dto.common.update.ListUpdateActionSet",
                            }
                        ],
                        "@type": "as.dto.common.update.IdListUpdateValue",
                    }
                else:
                    # handle single attributes (space, experiment, project, container, etc.)
                    value = self.__dict__.get("_" + attr, {})
                    if value is None:
                        pass
                    elif isinstance(value, bool):
                        if 'PluginUpdate' in up_obj['@type']:
                            # PluginUpdate is special
                            #TODO consolidate boolean attribute options
                            up_obj[attr] = {
                                "@type": "as.dto.common.update.FieldUpdateValue",
                                "isModified": True,
                                "value": value,
                            }
                        else:
                            # for boolean values where no type is needed
                            up_obj[attr] = value
                    elif isinstance(value, dict) and len(value) == 0:
                        # value is {}: it means that we want this attribute to be
                        # deleted, not updated.
                        up_obj[attr2ids[attr]] = {
                            "@type": "as.dto.common.update.FieldUpdateValue",
                            "isModified": True,
                        }
                    elif "isModified" in value and value["isModified"] == True:
                        val = {}
                        for x in ["identifier", "permId", "@type"]:
                            if x in value:
                                val[x] = value[x]

                        up_obj[attr2ids[attr]] = {
                            "@type": "as.dto.common.update.FieldUpdateValue",
                            "isModified": True,
                            "value": val,
                        }

        # update an existing entity
        if method_name is None:
            method_name = get_method_for_entity(self.entity, "update")
        request = {"method": method_name, "params": [self.openbis.token, [up_obj]]}
        return request

    def __getattr__(self, name):
        """handles all attribute requests dynamically.
        Values are returned in a sensible way, for example:
            the identifiers of parents, children and components are returned as an
            array of values, whereas attachments, users (of groups) and
            roleAssignments are returned as an array of dictionaries.
        """

        name_map = {
            "group": "authorizationGroup",
            "roles": "roleAssignments",
            "permid": "permId",
            "collection": "experiment",
            "object": "sample",
        }
        if name in name_map:
            name = name_map[name]

        int_name = "_" + name
        if int_name in self.__dict__:
            if int_name == "_attachments":
                attachments = []
                for att in self._attachments:
                    attachments.append(
                        {
                            "fileName": att.get("fileName"),
                            "title": att.get("title"),
                            "description": att.get("description"),
                            "version": att.get("version"),
                        }
                    )
                return attachments

            elif int_name == "_users":
                users = []
                for user in self._users:
                    users.append(
                        {
                            "firstName": user.get("firstName"),
                            "lastName": user.get("lastName"),
                            "email": user.get("email"),
                            "userId": user.get("userId"),
                            "space": user.get("space").get("code")
                            if user.get("space") is not None
                            else None,
                        }
                    )
                return users

            elif int_name == "_roleAssignments":
                ras = []
                for ra in self._roleAssignments:
                    ras.append(
                        {
                            "techId": ra.get("id").get("techId"),
                            "role": ra.get("role"),
                            "roleLevel": ra.get("roleLevel"),
                            "space": ra.get("space").get("code"),
                            "project": ra.get("role"),
                        }
                    )
                return ras

            elif int_name == '_parents':
                result = None
                if "_fetchOptions" in self.__dict__ and self.__dict__["_fetchOptions"] is not None:
                    if 'parents' in self.__dict__["_fetchOptions"] and self.__dict__["_fetchOptions"][
                        'parents'] is None:
                        return '--NOT FETCHED--'
                    result = self.__dict__[int_name]
                else:
                    result = self.__dict__[int_name]
                if isinstance(result, list):
                    values = []
                    for item in result:
                        if "identifier" in item:
                            values.append(item["identifier"])
                        elif "permId" in item:
                            values.append(item["permId"])
                        else:
                            values.append(item)
                    return values
                return result

            elif int_name == '_children':
                result = None
                if "_fetchOptions" in self.__dict__ and self.__dict__["_fetchOptions"] is not None:
                    if 'children' in self.__dict__["_fetchOptions"] and self.__dict__["_fetchOptions"][
                        'children'] is None:
                        return '--NOT FETCHED--'
                    result = self.__dict__[int_name]
                else:
                    result = self.__dict__[int_name]
                if isinstance(result, list):
                    values = []
                    for item in result:
                        if "identifier" in item:
                            values.append(item["identifier"])
                        elif "permId" in item:
                            values.append(item["permId"])
                        else:
                            values.append(item)
                    return values
                return result

            # if the attribute contains a list,
            # return a list of either identifiers, codes or
            # permIds (whatever is available first)
            elif isinstance(self.__dict__[int_name], list):
                values = []
                for item in self.__dict__[int_name]:
                    if "identifier" in item:
                        values.append(item["identifier"])
                    elif "code" in item:
                        values.append(item["code"])
                    elif "userId" in item:
                        values.append(item["userId"])
                    elif "permId" in item:
                        values.append(item["permId"])
                    else:
                        values.append(item)
                return values

            # attribute contains a dictionary: same procedure as above.
            elif isinstance(self.__dict__[int_name], dict):
                if "identifier" in self.__dict__[int_name]:
                    return self.__dict__[int_name]["identifier"]
                elif "code" in self.__dict__[int_name]:
                    return self.__dict__[int_name]["code"]
                elif "name" in self.__dict__[int_name]:
                    return self.__dict__[int_name]["name"]
                elif "userId" in self.__dict__[int_name]:
                    return self.__dict__[int_name]["userId"]
                elif "permId" in self.__dict__[int_name]:
                    return self.__dict__[int_name]["permId"]
                elif "id" in self.__dict__[int_name]:
                    return self.__dict__[int_name]["id"]
                else:
                    return self.__dict__[int_name]

            else:
                return self.__dict__[int_name]
        else:
            return None

    def __setattr__(self, name, value):
        """This method is always invoked whenever we assign an attribute to an
        object, e.g.
            new_sample.space = 'MATERIALS'
            new_sample.parents = ['/MATERIALS/YEAST747']
        """

        # experiment aka collection, sample aka object
        name_map = {
            "collection": "experiment",
            "object": "sample",
            "parent": "parents",
            "child": "children",
        }
        if name in name_map:
            name = name_map[name]

        if self._is_new:
            if name not in self._defs["attrs_new"]:
                raise ValueError(
                    f'No such attribute: «{name}» for entity: {self.entity}. Allowed attributes are: {self._defs["attrs_new"]}'
                )
        else:
            if name not in self._defs["attrs_up"]:
                raise ValueError(
                    f'No such attribute: «{name}» for entity: {self.entity}. Allowed attributes are: {self._defs["attrs_up"]}'
                )

        if name in ["parents", "children", "components"]:

            if not isinstance(value, list):
                value = [value]
            objs = []
            for val in value:
                if isinstance(val, str):
                    # fetch objects in openBIS, make sure they actually exists
                    obj = getattr(self._openbis, "get_" + self._entity.lower())(val)
                    objs.append(obj)
                elif getattr(val, "_permId"):
                    # we got an existing object
                    objs.append(val)

            permids = []
            for item in objs:
                if getattr(item, "_identifier") is not None:
                    id = item._identifier
                elif getattr(item, "_permId") is not None:
                    id = item._permId
                else:
                    return

                # remove any existing @id keys to prevent jackson parser errors
                id.pop("@id", None)

                permids.append(id)

            self.__dict__["_" + name] = permids
        elif name in ["tags"]:
            self.set_tags(value)

        elif name in ["users"]:
            self.set_users(value)

        elif name in ["vocabulary"]:
            if value is None or value == "":
                self.__dict__["_vocabulary"] = None
            else:
                self.__dict__["_vocabulary"] = {
                    "@type": "as.dto.vocabulary.id.VocabularyPermId",
                    "permId": value.upper(),
                }

        elif name in ["sampleType"]:
            if value is None or value == "":
                self.__dict__["_sampleType"] = None
            else:
                self.__dict__["_sampleType"] = {
                    "@type": "as.dto.entitytype.id.EntityTypePermId",
                    "permId": value.upper(),
                    "entityKind": "SAMPLE",
                }

        elif name in ["validationPlugin"]:
            if value is None or value == "":
                self.__dict__["_validationPlugin"] = None
            else:
                self.__dict__["_validationPlugin"] = {
                    "@type": "as.dto.plugin.id.PluginPermId",
                    "permId": value,
                }

        elif name in ["materialType"]:
            if value is None or value == "":
                self.__dict__["_materialType"] = None
            else:
                self.__dict__["_materialType"] = {
                    "@type": "as.dto.entitytype.id.EntityTypePermId",
                    "permId": value.upper(),
                    "entityKind": "MATERIAL",
                }

        elif name in ["attachments"]:
            if isinstance(value, list):
                for item in value:
                    if isinstance(item, dict):
                        self.add_attachment(**item)
                    else:
                        self.add_attachment(item)

            else:
                self.add_attachment(value)

        elif name in ["space"]:
            obj = None
            if value is None:
                self.__dict__["_" + name] = None
                return

            if isinstance(value, str):
                # fetch object in openBIS, make sure it actually exists
                obj = getattr(self._openbis, "get_" + name)(value)
            else:
                obj = value

            self.__dict__["_" + name] = obj.data["permId"]

            # mark attribute as modified, if it's an existing entity
            if self.is_new:
                pass
            else:
                self.__dict__["_" + name]["isModified"] = True

        elif name in ["sample", "experiment", "project"]:
            obj = None
            if isinstance(value, str):
                # fetch object in openBIS, make sure it actually exists
                obj = getattr(self._openbis, "get_" + name)(value)
            elif value is None:
                self.__dict__["_" + name] = {}
                return
            else:
                obj = value

            self.__dict__["_" + name] = obj.data["identifier"]

            # mark attribute as modified, if it's an existing entity
            if self.is_new:
                pass
            else:
                self.__dict__["_" + name]["isModified"] = True

        elif name in ["identifier"]:
            raise KeyError(f"you can not modify the {name}")
        elif name == "code":
            try:
                if self._type["autoGeneratedCode"]:
                    raise KeyError(
                        f"This {self.entity}Type has auto-generated code. You cannot set a code"
                    )
            except KeyError:
                pass
            except TypeError:
                pass

            self.__dict__["_code"] = value

        elif name in ["userId", "description"]:
            # values that are directly assigned
            self.__dict__["_" + name] = value

        elif name in ["userIds"]:
            self.add_users(value)

        elif name in self._defs:
            # enum: check whether value is a valid constant
            if value and value not in self._defs[name]:
                raise ValueError(
                    f"Allowed values for enum {name} are: {self._defs[name]}"
                )
            else:
                self.__dict__["_" + name] = value
        else:
            self.__dict__["_" + name] = value

    def get_type(self):
        return self._type

    def _ident_for_whatever(self, whatever):
        if isinstance(whatever, str):
            # fetch parent in openBIS, we are given an identifier
            obj = getattr(self._openbis, "get_" + self._entity.lower())(whatever)
        else:
            # we assume we got an object
            obj = whatever

        ident = None
        if getattr(obj, "_identifier"):
            ident = obj._identifier
        elif getattr(obj, "_permId"):
            ident = obj._permId

        ident.pop("@id", None)
        return ident

    def get_container(self, **kwargs):
        return getattr(self._openbis, "get_" + self._entity.lower())(
            self.container, **kwargs
        )

    def get_containers(self, **kwargs):
        """get the containers and return them as a list (Things/DataFrame)
        or return empty list
        """
        return getattr(self._openbis, "get_" + self._entity.lower())(
            self.containers, **kwargs
        )

    def set_containers(self, containers_to_set):
        """set the new _containers list"""
        self.__dict__["_containers"] = []
        self.add_containers(containers_to_set)

    def add_containers(self, containers_to_add):
        """add component to _containers list"""
        if not isinstance(containers_to_add, list):
            containers_to_add = [containers_to_add]
        for component in containers_to_add:
            ident = self._ident_for_whatever(component)
            if ident not in self.__dict__["_containers"]:
                self.__dict__["_containers"].append(ident)

    def del_containers(self, containers_to_remove):
        """remove component from _containers list"""
        if not isinstance(containers_to_remove, list):
            containers_to_remove = [containers_to_remove]
        for component in containers_to_remove:
            ident = self._ident_for_whatever(component)
            for i, item in enumerate(self.__dict__["_containers"]):
                if (
                    "identifier" in ident
                    and "identifier" in item
                    and ident["identifier"] == item["identifier"]
                ):
                    self.__dict__["_containers"].pop(i, None)
                elif (
                    "permId" in ident
                    and "permId" in item
                    and ident["permId"] == item["permId"]
                ):
                    self.__dict__["_containers"].pop(i, None)

    def get_components(self, **kwargs):
        """Samples and DataSets may contain other DataSets and Samples. This function returns the
        contained Samples/DataSets (a.k.a. components) as a list (Things/DataFrame)
        """
        return getattr(self._openbis, "get_" + self._entity.lower())(
            self.components, **kwargs
        )

    get_contained = get_components  # Alias

    def set_components(self, components_to_set):
        """Samples and DataSets may contain other DataSets and Samples. This function sets the
        contained Samples/DataSets (a.k.a. components)
        """
        self.__dict__["_components"] = []
        self.add_components(components_to_set)

    set_contained = set_components  # Alias

    def add_components(self, components_to_add):
        """Samples and DataSets may contain other DataSets and Samples. This function adds
        additional Samples/DataSets to the current object.
        """
        if not isinstance(components_to_add, list):
            components_to_add = [components_to_add]
        for component in components_to_add:
            ident = self._ident_for_whatever(component)
            if ident not in self.__dict__["_components"]:
                self.__dict__["_components"].append(ident)

    add_contained = add_components  # Alias

    def del_components(self, components_to_remove):
        """Samples and DataSets may contain other DataSets and Samples. This function removes
        additional Samples/DataSets from the current object.
        """
        if not isinstance(components_to_remove, list):
            components_to_remove = [components_to_remove]
        for component in components_to_remove:
            ident = self._ident_for_whatever(component)
            for i, item in enumerate(self.__dict__["_components"]):
                if (
                    "identifier" in ident
                    and "identifier" in item
                    and ident["identifier"] == item["identifier"]
                ):
                    self.__dict__["_components"].pop(i, None)
                elif (
                    "permId" in ident
                    and "permId" in item
                    and ident["permId"] == item["permId"]
                ):
                    self.__dict__["_components"].pop(i, None)

    del_contained = del_components  # Alias

    def get_parents(self, **kwargs):
        """get the current parents and return them as a list (Things/DataFrame)
        or return empty list
        """
        parents = getattr(self.openbis, "get_" + self.entity.lower() + "s")(
            withChildren=self.identifier, **kwargs
        )
        if self.__dict__["_fetchOptions"] is None:
            self.__dict__["_fetchOptions"] = {}
        if 'parents' not in self.__dict__["_fetchOptions"] or self.__dict__["_fetchOptions"]['parents'] is None:
            self.__dict__["_fetchOptions"]['parents'] = {}
        if len(parents) > 0:
            self.__dict__["_parents"] = list(parents[['identifier']].values.flatten())
        else:
            self.__dict__["_parents"] = []
        if self.__dict__["_parents"] is None or self.__dict__["_parents"] == []:
            return []
        return getattr(self._openbis, "get_" + self._entity.lower())(
            self.parents, **kwargs
        )

    def set_parents(self, parents_to_set):
        """set the new _parents list"""
        self.__dict__["_parents"] = []
        self.add_parents(parents_to_set)

    def add_parents(self, parents_to_add):
        """add parent to _parents list"""
        if not isinstance(parents_to_add, list):
            parents_to_add = [parents_to_add]
        if self.__dict__["_parents"] is None:
            self.__dict__["_parents"] = []
        for parent in parents_to_add:
            ident = self._ident_for_whatever(parent)
            if ident not in self.__dict__["_parents"]:
                self.__dict__["_parents"].append(ident)

    def del_parents(self, parents_to_remove):
        """remove parent from _parents list"""
        if not isinstance(parents_to_remove, list):
            parents_to_remove = [parents_to_remove]
        if self.__dict__["_parents"] is None:
            self.__dict__["_parents"] = []
        for parent in parents_to_remove:
            ident = self._ident_for_whatever(parent)
            for i, item in enumerate(self.__dict__["_parents"]):
                if (
                    "identifier" in ident
                    and "identifier" in item
                    and ident["identifier"] == item["identifier"]
                ):
                    self.__dict__["_parents"].pop(i)
                elif (
                    "permId" in ident
                    and "permId" in item
                    and ident["permId"] == item["permId"]
                ):
                    self.__dict__["_parents"].pop(i)

    def get_children(self, **kwargs):
        """get the current children and return them as a list (Things/DataFrame)
        or return empty list
        """
        children = getattr(self.openbis, "get_" + self.entity.lower() + "s")(
            withParents=self.identifier, **kwargs
        )
        if self.__dict__["_fetchOptions"] is None:
            self.__dict__["_fetchOptions"] = {}
        if 'children' not in self.__dict__["_fetchOptions"] or self.__dict__["_fetchOptions"]['children'] is None:
            self.__dict__["_fetchOptions"]['children'] = {}
        if len(children) > 0:
            self.__dict__["_children"] = list(children[['identifier']].values.flatten())
        else:
            self.__dict__["_children"] = []
        if self.__dict__["_children"] is None or self.__dict__["_children"] == []:
            return []
        return getattr(self._openbis, "get_" + self._entity.lower())(
            self.children, **kwargs
        )

    def set_children(self, children_to_set):
        """set the new _children list"""
        self.__dict__["_children"] = []
        self.add_children(children_to_set)

    def add_children(self, children):
        """add children to _children list"""
        if getattr(self, "_children") is None:
            self.__dict__["_children"] = []
        if self.__dict__["_children"] is None:
            self.__dict__["_children"] = []
        if not isinstance(children, list):
            children = [children]
        for child in children:
            self.__dict__["_children"].append(self._ident_for_whatever(child))

    def del_children(self, children):
        """remove children from _children list"""
        if getattr(self, "_children") is None:
            return
        if not isinstance(children, list):
            children = [children]
        if self.__dict__["_children"] is None:
            self.__dict__["_children"] = []
        for child in children:
            ident = self._ident_for_whatever(child)
            for i, item in enumerate(self.__dict__["_children"]):
                if (
                    "identifier" in ident
                    and "identifier" in item
                    and ident["identifier"] == item["identifier"]
                ):
                    self.__dict__["_children"].pop(i)
                elif (
                    "permId" in ident
                    and "permId" in item
                    and ident["permId"] == item["permId"]
                ):
                    self.__dict__["_children"].pop(i)

    @property
    def tags(self):
        if getattr(self, "_tags") is not None:
            return [x["code"] for x in self._tags]

    def get_tags(self):
        if getattr(self, "_tags") is not None:
            return self._openbis.get_tag([x["permId"] for x in self._tags])

    def set_tags(self, tags):
        """set _tags list"""

        self.__dict__["_tags"] = []
        self.add_tags(tags)

    def add_tags(self, tags):
        """add tags to _tags list"""
        if not isinstance(tags, list):
            tags = [tags]

        for tag in tags:
            if isinstance(tag, str):
                tag_obj = self._openbis.get_tag(tag)
                tag_dict = {
                    "code": tag_obj.code,
                    "permId": tag_obj.permId,
                }
            else:
                tag_dict = {
                    "code": tag["code"],
                    "permId": tag["permId"]["permId"],
                }

            if tag_dict not in self.__dict__["_tags"]:
                self.__dict__["_tags"].append(tag_dict)

    def del_tags(self, tags):
        """remove tags from _tags list"""
        if not isinstance(tags, list):
            tags = [tags]

        for tag in tags:
            for i, tag_dict in enumerate(self.__dict__["_tags"]):
                if tag in tag_dict[i]["code"] or tag in tag_dict[i]["permId"]:
                    tag_dict.pop(i, None)

    def set_users(self, userIds):
        if userIds is None:
            return
        if getattr(self, "_userIds") is None:
            self.__dict__["_userIds"] = []
        if not isinstance(userIds, list):
            userIds = [userIds]
        for userId in userIds:
            self.__dict__["_userIds"].append(
                {"permId": userId, "@type": "as.dto.person.id.PersonPermId"}
            )

    def add_users(self, userIds):
        if userIds is None:
            return
        if getattr(self, "_changed_users") is None:
            self.__dict__["_changed_users"] = {}

        if not isinstance(userIds, list):
            userIds = [userIds]
        for userId in userIds:
            self.__dict__["_changed_users"][userId] = {"action": "Add"}

    add_members = add_users  # Alias

    def del_users(self, userIds):
        if userIds is None:
            return
        if getattr(self, "_changed_users") is None:
            self.__dict__["_changed_users"] = {}

        if not isinstance(userIds, list):
            userIds = [userIds]
        for userId in userIds:
            self.__dict__["_changed_users"][userId] = {"action": "Remove"}

    del_members = del_users  # Alias

    def get_attachments(self):
        if getattr(self, "_attachments") is None:
            return None
        else:
            return DataFrame(self._attachments)[
                ["fileName", "title", "description", "version"]
            ]

    def add_attachment(self, fileName, title=None, description=None):
        att = Attachment(filename=fileName, title=title, description=description)
        if getattr(self, "_attachments") is None:
            self.__dict__["_attachments"] = []
        self._attachments.append(att.get_data_short())

        if getattr(self, "_new_attachments") is None:
            self.__dict__["_new_attachments"] = []
        self._new_attachments.append(att)

    def download_attachments(self, destination_folder=None):
        method = "get" + self.entity.lower().capitalize() + "s"
        request = {
            "method": method,
            "params": [
                self._openbis.token,
                [self._permId],
                {
                    **fetch_option[self.entity.lower()],
                    "attachments": fetch_option["attachmentsWithContent"],
                },
            ],
        }
        resp = self._openbis._post_request(self._openbis.as_v3, request)
        attachments = resp[self.permId]["attachments"]
        file_list = []

        if destination_folder is None:
            destination_folder = os.getcwd()
        for attachment in attachments:
            filename = (
                pathlib.Path(destination_folder)
                / pathlib.Path(attachment["fileName"]).name
            )
            os.makedirs(os.path.dirname(filename), exist_ok=True)
            with open(filename, "wb") as att:
                content = base64.b64decode(attachment["content"])
                att.write(content)
            file_list.append(str(filename))
        return file_list

    def all(self):
        """Return all attributes of an entity in a dict"""
        attrs = {}
        for attr in self._defs["attrs"]:
            if attr == "attachments":
                continue
            attrs[attr] = getattr(self, attr)
        return attrs

    def _repr_html_(self):
        def nvl(val, string=""):
            if val is None:
                return string
            return val

        html = """
            <table border="1" class="dataframe">
            <thead>
                <tr style="text-align: right;">
                <th>attribute</th>
                <th>value</th>
                </tr>
            </thead>
            <tbody>
        """

        attrs = self._defs["attrs_new"] if self.is_new else self._defs["attrs"]
        for attr in attrs:
            if attr == "attachments":
                continue
            html += f'<tr> <td>{attr}</td> <td>{nvl(getattr(self, attr, ""), "")}</td> </tr>'
        if getattr(self, "_attachments") is not None:
            html += "<tr><td>attachments</td><td>"
            html += "<br/>".join(att["fileName"] for att in self._attachments)
            html += "</td></tr>"

        html += """
            </tbody>
            </table>
        """
        return html

    def __repr__(self):
        """When using IPython, this method displays a nice table
        of all attributes and their values when the object is printed.
        """

        headers = ["attribute", "value"]
        lines = []
        attrs = self._defs["attrs_new"] if self.is_new else self._defs["attrs"]
        for attr in attrs:
            if attr == "attachments":
                continue
            elif attr == "users" and "_users" in self.__dict__:
                lines.append([attr, ", ".join(att["userId"] for att in self._users)])
            elif attr == "roleAssignments" and "_roleAssignments" in self.__dict__:
                roles = []
                for role in self._roleAssignments:
                    if role.get("space") is not None:
                        roles.append(
                            f'{role.get("role")} ({role.get("space").get("code")})'
                        )
                    else:
                        roles.append(role.get("role"))

                lines.append([attr, ", ".join(roles)])

            else:
                lines.append([attr, nvl(getattr(self, attr, ""))])
        return tabulate(lines, headers=headers)



#   Copyright ETH 2018 - 2024 Zürich, Scientific IT Services
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#        http://www.apache.org/licenses/LICENSE-2.0
#   
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
from .attribute import AttrHolder
from .openbis_object import OpenBisObject
from .utils import VERBOSE


class RoleAssignment(OpenBisObject):
    """managing openBIS role assignments"""

    def __init__(self, openbis_obj, data=None, **kwargs):
        self.__dict__["openbis"] = openbis_obj
        self.__dict__["a"] = AttrHolder(openbis_obj, "roleAssignment")

        if data is not None:
            self.a(data)
            self.__dict__["data"] = data

        if kwargs is not None:
            for key in kwargs:
                setattr(self, key, kwargs[key])

    def __dir__(self):
        """all the available methods and attributes that should be displayed
        when using the autocompletion feature (TAB) in Jupyter
        """
        return ["id", "role", "roleLevel", "space", "project", "group"]

    def __str__(self):
        return f"{self.get('role')}"

    def delete(self, reason="no reason specified"):
        self.openbis.delete_openbis_entity(
            entity="roleAssignment", objectId=self._id, reason=reason
        )
        if VERBOSE:
            print(
                f"RoleAssignment role={self.role}, roleLevel={self.roleLevel} successfully deleted."
            )



#   Copyright ETH 2024 -2025 Zürich, Scientific IT Services
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

import json
import copy

from pandas import DataFrame

def _nonzero(num):
    if num != 0:
        return 1
    return 0

def _get_headers(count):
    """Algorithm for generating headers, maximum number of columns supported: 26*26=676"""
    if count < 1:
        raise ValueError("Can not create spreadsheet without columns!")
    min_char = ord('A')
    alphabet_max = 26
    headers = [chr(x) for x in range(min_char, min_char+min(alphabet_max, count))]
    if count > alphabet_max:
        for x in range(count // alphabet_max):
            char = min_char + x
            headers += [chr(char) + chr(min_char+y) for y in range(min(alphabet_max, count - alphabet_max*(x+1)))]
    return headers

class Spreadsheet:
    headers: list
    data: list
    style: dict
    meta: dict
    width: list
    values: list
    version: str

    def __init__(self, columns=10, rows=10):
        self.version = "1"
        self.headers = _get_headers(columns)
        self.data = [["" for _ in range(columns)] for _ in range(rows)]
        self.style = {
                header + str(y): "text-align: center;" for header in self.headers for y in range(1, rows+1)
            }
        self.meta = {}
        self.width = [50 for _ in range(columns)]
        self.values = [["" for _ in range(columns)] for _ in range(rows)]

    def _get_index_str(self, index):
        index = index.strip()
        column = ""
        row = ""
        headers = True
        for i in index:
            if i.isalpha():
                if not headers:
                    raise ValueError("Wrong index schema!")
                column += i
            elif ord(i) >= 48 and ord(i) <= 57:
                headers = False
                row += i
            else:
                raise ValueError("Wrong index schema!")
        if not column in self.headers:
            raise ValueError(f"Column '{column}' does not exists!")
        if row == "":
            raise ValueError("Missing row index!")
        row = int(row)
        if len(self.data) < row or row < 1:
            raise ValueError(f"Row '{row}' does not exists!")

        return row - 1, self.headers.index(column)

    def _get_index_tuple(self, index):
        column, row = index
        if column is None or column == "" or row is None or row == "":
            raise ValueError(f"'{index}' is not a valid index!")

        if str(column).isdigit():
            column = int(column)
        else:
            if not column in self.headers:
                raise ValueError(f"Column '{column}' does not exists!")
            column = self.headers.index(column) + 1

        row = int(row)
        if len(self.data) < row or row < 1:
            raise ValueError(f"Row '{row}' does not exists!")

        return row, column

    def _get_index(self, index):
        if index is None:
            raise ValueError("Index must not be None!")
        if isinstance(index, tuple):
            return self._get_index_tuple(index)
        raise ValueError(f"'{index}' is not a valid index!")

    def __getitem__(self, index):
        (row, column) = self._get_index(index)
        return self.CellBuilder(self, column, row)

    def get_column_count(self):
        return len(self.headers)

    def get_row_count(self):
        return len(self.data)

    def get_version(self):
        return self.version

    def get_meta_data(self):
        return self.meta

    def get_formulas(self):
        return copy.deepcopy(self.data)

    def get_headers(self):
        return copy.deepcopy(self.headers)

    def get_values(self):
        return copy.deepcopy(self.values)

    def get_width(self):
        return copy.deepcopy(self.width)

    def get_style(self):
        return copy.deepcopy(self.style)

    def df(self, attribute):
        options = ['headers', 'formulas', 'width', 'values']
        if attribute not in options:
            raise ValueError(f"Attribute '{attribute}' not found in the spreadsheet! Available attributes are: {options}")

        if attribute == 'headers':
            return DataFrame(self.headers)
        elif attribute == 'formulas':
            return DataFrame(self.data, columns=self.headers, index=range(1, len(self.data)+1, 1))
        elif attribute == 'values':
            return DataFrame(self.values, columns=self.headers, index=range(1, len(self.values) + 1, 1))
        elif attribute == 'width':
            return DataFrame(self.width)


    def add_column(self, column_name=None):
        if column_name is None:
            column_name = _get_headers(len(self.headers)+1)[-1]
        self.headers += [column_name]
        for row in self.data:
            row += ['']
        for row in self.values:
            row += ['']
        self.width += [50]
        for x in range(1, len(self.data[0])):
            self.style[column_name + str(x)] =  "text-align: center;"

    def add_row(self):
        self.data += [['' for _ in range(len(self.headers))]]
        self.values += [['' for _ in range(len(self.headers))]]

        for header in self.headers:
            self.style[header + str(len(self.data))] =  "text-align: center;"

    def delete_row(self, row_number):
        row = int(row_number)
        if len(self.data) < row or row < 1:
            raise ValueError(f"Row '{row}' does not exists!")

        for header in self.headers:
            for i in range(row, len(self.data), 1):
                self.style[header + str(i)] = self.style[header + str(i+1)]
            del self.style[header + str(len(self.data))]

        self.data.pop(row - 1)
        self.values.pop(row - 1)

    def delete_column(self, column_identifier):
        if str(column_identifier).isdigit():
            column_identifier = int(column_identifier)
            if column_identifier < 1 or column_identifier > len(self.headers):
                raise ValueError(f"Column '{column_identifier}' does not exists!")
            column_index = column_identifier-1
            column_label = self.headers[column_index]
        else:
            if not column_identifier in self.headers:
                raise ValueError(f"Column '{column_identifier}' does not exists!")
            column_index = self.headers.index(column_identifier)
            column_label = self.headers[column_index]

        self.headers.pop(column_index)
        self.width.pop(column_index)
        for i in range(len(self.data)):
            self.data[i].pop(column_index)
            self.values[i].pop(column_index)
            del self.style[column_label+str(i+1)]


    def cell(self, column_identifier, row_number):
        return self.CellBuilder(self, column_identifier, row_number)

    def column(self, column_identifier):
        return self.ColumnBuilder(self, column_identifier)

    class CellBuilder:

        def __init__(self, parent, column_identifier, row_number):
            self.__dict__['parent'] = parent
            if column_identifier is None or column_identifier == "" or row_number is None or row_number == "":
                raise ValueError(f"('{column_identifier}','{row_number}') is not a valid index!")

            if str(column_identifier).isdigit():
                column_identifier = int(column_identifier)
                if column_identifier < 1 or column_identifier > len(parent.headers):
                    raise ValueError(f"Column '{column_identifier}' does not exists!")
                self.__dict__['column_index'] = column_identifier - 1
                self.__dict__['column_label'] = parent.headers[self.__dict__['column_index']]
            else:
                if not column_identifier in parent.headers:
                    raise ValueError(f"Column '{column_identifier}' does not exists!")
                self.__dict__['column_index'] = parent.headers.index(column_identifier)
                self.__dict__['column_label'] = parent.headers[self.__dict__['column_index']]

            row = int(row_number)
            if len(parent.data) < row or row < 1:
                raise ValueError(f"Row '{row}' does not exists!")
            self.__dict__['row_index'] = row - 1

        def __getattr__(self, name):
            row = self.__dict__['row_index']
            column = self.__dict__['column_index']
            label = self.__dict__['column_label']
            if name == "formula":
                return self.__dict__['parent'].data[row][column]
            elif name == "value":
                return self.__dict__['parent'].values[row][column]
            elif name == "style":
                return self.__dict__['parent'].style[label + str(row+1)]
            elif name == "column_header":
                return label
            elif name == "column_number":
                return column+1
            elif name == "row_number":
                return row+1
            else:
                raise ValueError(f"No such attribute '{name}' found!")


        def __setattr__(self, name, value):
            row = self.__dict__['row_index']
            column = self.__dict__['column_index']
            label = self.__dict__['column_label']
            if name == "formula":
                self.__dict__['parent'].data[row][column] = value
            elif name == "value":
                self.__dict__['parent'].values[row][column] = value
            elif name == "style":
                self.__dict__['parent'].style[label + str(column + 1)]  = value
            else:
                raise ValueError(f"No such attribute '{name}' found!")

        def __str__(self):
            attr = self.__dict__
            row = attr['row_index']
            column = attr['column_index']
            return f"Cell[column={column}, row={row}, formula={attr['parent'].data[row][column]}, value={attr['parent'].values[row][column]}]"


    class ColumnBuilder:
        def __init__(self, parent, column_identifier):
            # self.__dict__['__df'] = None
            self.__dict__['parent'] = parent
            if column_identifier is None or column_identifier == "":
                raise ValueError(f"('{column_identifier}') is not a valid column index!")

            if str(column_identifier).isdigit():
                column_identifier = int(column_identifier)
                if column_identifier < 1 or column_identifier > len(parent.headers):
                    raise ValueError(f"Column '{column_identifier}' does not exists!")
                self.__dict__['column_index'] = column_identifier - 1
                self.__dict__['column_label'] = parent.headers[self.__dict__['column_index']]
            else:
                if not column_identifier in parent.headers:
                    raise ValueError(f"Column '{column_identifier}' does not exists!")
                self.__dict__['column_index'] = parent.headers.index(column_identifier)
                self.__dict__['column_label'] = parent.headers[self.__dict__['column_index']]

        def __getattr__(self, name):
            if name == "header":
                return self.__dict__['column_label']
            elif name == "width":
                return self.__dict__['parent'].width[self.__dict__['column_index']]
            elif name == "column_number":
                return self.__dict__['column_index']+1
            else:
                raise ValueError(f"No such attribute '{name}' found!")

        def __setattr__(self, name, value):
            if name == "header":
                self.__dict__['parent'].headers[self.__dict__['column_index']] = value
            elif name == "width":
                self.__dict__['parent'].width[self.__dict__['column_index']] = value
            else:
                raise ValueError(f"No such attribute '{name}' found!")

        def __str__(self):
            attr = self.__dict__
            return f"Column[column={attr['column_index']}, header={attr['column_label']}]"


    def __str__(self):
        return json.dumps(self.__dict__, default=lambda x: x.__dict__)

    def __repr__(self):
        return json.dumps(self.__dict__, default=lambda x: x.__dict__)

    def to_json(self):

        def dictionary_creator(x):
            dictionary = x.__dict__
            return dictionary

        return json.dumps(self, default=dictionary_creator, sort_keys=True, indent=4)

    @classmethod
    def from_dict(cls, data):
        if data is None:
            return None
        result = cls(10)
        for prop in cls.__annotations__.keys():
            attribute = data.get(prop)
            result.__dict__[prop] = attribute
        return result



#   Copyright ETH 2018 - 2023 Zürich, Scientific IT Services
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#        http://www.apache.org/licenses/LICENSE-2.0
#   
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
import site
import sys

site.ENABLE_USER_SITE = "--user" in sys.argv[1:]
if sys.version_info < (3, 6):
    sys.exit("Sorry, Python < 3.6 is not supported")

from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="pybis",
    version="6.7.0.0-rc0",
    author="ID SIS • ETH Zürich",
    author_email="openbis-support@id.ethz.ch",
    description="openBIS connection and interaction, optimized for using with Jupyter",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://sissource.ethz.ch/sispub/openbis/-/tree/master/api-openbis-python3-pybis",
    packages=find_packages(),
    license="Apache Software License Version 2.0",
    install_requires=[
        "pytest",
        "requests",
        "urllib3",
        "pandas",
        "texttable",
        "tabulate",
        "python-dateutil",
    ],
    project_urls={
        "Documentation": "https://openbis.readthedocs.io/en/latest/software-developer-documentation/apis/python-v3-api.html",
        "Changelog": "https://sissource.ethz.ch/sispub/openbis/-/blob/master/api-openbis-python3-pybis/src/python/CHANGELOG.md"
    },
    python_requires=">=3.6",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
)



#   Copyright ETH 2018 - 2024 Zürich, Scientific IT Services
# 
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
# 
#        http://www.apache.org/licenses/LICENSE-2.0
#   
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
import json
import os
import random
import time
import urllib.parse
import uuid
import zipfile
from functools import partialmethod
from pathlib import Path
from queue import Queue
from threading import Thread
from typing import Set, Optional, List
from urllib.parse import urljoin, quote

import requests
from pandas import DataFrame
from requests import Session
from requests.adapters import HTTPAdapter, Retry
from tabulate import tabulate

from .definitions import (
    openbis_definitions,
    get_type_for_entity,
    get_fetchoption_for_entity,
)
from .fast_download import FastDownload
from .openbis_object import OpenBisObject
from .things import Things
from .utils import (
    VERBOSE,
    parse_jackson,
    extract_permid,
    extract_code,
    extract_downloadUrl,
)

# needed for Data upload
PYBIS_PLUGIN = "dataset-uploader-api"
DATASET_DEFINITIONS = openbis_definitions("dataSet")
DSS_ENDPOINT = "/datastore_server/rmi-data-store-server-v3.json"
SESSION_WORKSPACE = "/datastore_server/session_workspace_file_upload"
REQUEST_RETRIES_COUNT = 5


def signed_to_unsigned(sig_int):
    """openBIS delivers crc32 checksums as signed integers.
    If the number is negative, we just have to add 2**32
    We display the hex number to match with the classic UI
    """
    if sig_int < 0:
        sig_int += 2 ** 32
    return "%x" % (sig_int & 0xFFFFFFFF)


class DataSet(
    OpenBisObject,
    entity="dataSet",
    single_item_method_name="get_dataset",
):
    """DataSet are openBIS objects that contain the actual files."""

    def __init__(
            self,
            openbis_obj,
            type,
            data=None,
            files=None,
            zipfile=None,
            folder=None,
            kind=None,
            props=None,
            **kwargs,
    ):

        if kind == "PHYSICAL":
            if files is None and zipfile is None:
                raise ValueError("please provide at least one file")

            if files and zipfile:
                raise ValueError(
                    "please provide either a list of files or a single zipfile"
                )

            if zipfile:
                files = [zipfile]
                self.__dict__["isZipDirectoryUpload"] = True
            else:
                self.__dict__["isZipDirectoryUpload"] = False

            if files:
                if isinstance(files, str):
                    files = [files]

                for file in files:
                    if not os.path.exists(file):
                        raise ValueError(f"File {file} does not exist")

                self.__dict__["files"] = files

        # initialize the OpenBisObject
        super().__init__(openbis_obj, type=type, data=data, props=props, **kwargs)

        self.__dict__["files_in_wsp"] = []

        # existing DataSet
        if data is not None:
            if data["physicalData"] is None:
                self.__dict__["shareId"] = None
                self.__dict__["location"] = None
            else:
                self.__dict__["shareId"] = data["physicalData"]["shareId"]
                self.__dict__["location"] = data["physicalData"]["location"]

        if kind is not None:
            kind = kind.upper()
            allowed_kinds = ["PHYSICAL", "CONTAINER", "LINK"]
            if kind not in allowed_kinds:
                raise ValueError(
                    f"only these values are allowed for kind: {allowed_kinds}"
                )
            self.a.__dict__["_kind"] = kind

        self.__dict__["folder"] = folder

        if getattr(self, "parents") is None:
            self.a.__dict__["_parents"] = []
        else:
            if not self.is_new:
                self.a.__dict__["_parents_orig"] = self.a.__dict__["_parents"]

        if getattr(self, "children") is None:
            self.a.__dict__["_children"] = []
        else:
            if not self.is_new:
                self.a.__dict__["_children_orig"] = self.a.__dict__["_children"]

        if getattr(self, "container") is None:
            self.a.__dict__["_container"] = []
        else:
            if not self.is_new:
                self.a.__dict__["_container_orig"] = self.a.__dict__["_container"]

        if getattr(self, "component") is None:
            self.a.__dict__["_component"] = []
        else:
            if not self.is_new:
                self.a.__dict__["_component_orig"] = self.a.__dict__["_component"]

    def __str__(self):
        return self.data["code"]

    def __dir__(self):
        return [
            "get_parents()",
            "get_children()",
            "get_components()",
            "get_contained()",
            "get_containers()",
            "add_parents()",
            "add_children()",
            "add_components()",
            "add_contained()",
            "add_containers()",
            "del_parents()",
            "del_children()",
            "del_components()",
            "del_contained()",
            "del_containers()",
            "set_parents()",
            "set_children()",
            "set_components()",
            "set_contained()",
            "set_containers()",
            "set_tags()",
            "add_tags()",
            "del_tags()",
            "add_attachment()",
            "get_attachments()",
            "download_attachments()",
            "get_files()",
            "file_list",
            "file_links",
            "rel_file_links",
            "physicalData",
            "download()",
            "download_path",
            "is_physical()",
            "symlink()",
            "is_symlink()",
            "archive()",
            "unarchive()",
            "save()",
            "delete()",
            "mark_to_be_deleted()",
            "unmark_to_be_deleted()",
            "is_marked_to_be_deleted()",
            "attrs",
            "props",
        ] + super().__dir__()

    def __setattr__(self, name, value):
        if name in ["folder"]:
            self.__dict__[name] = value
        elif name in ["p", "props"]:
            if isinstance(value, dict):
                for p in value:
                    setattr(self.__dict__["p"], p, value[p])
            else:
                raise ValueError("please provide a dictionary for setting properties")
        else:
            super(DataSet, self).__setattr__(name, value)

    @property
    def props(self):
        return self.__dict__["p"]

    @property
    def type(self):
        return self.__dict__["type"]

    @type.setter
    def type(self, type_name):
        dataset_type = self.openbis.get_dataset_type(type_name.upper())
        self.p.__dict__["_type"] = dataset_type
        self.a.__dict__["_type"] = dataset_type

    @property
    def physicalData(self):
        if "physicalData" in self.data:
            return PhysicalData(data=self.data["physicalData"])

    @property
    def linkedData(self):
        if "linkedData" in self.data:
            return LinkedData(data=self.data["linkedData"])

    @property
    def status(self):
        ds = self.openbis.get_dataset(self.permId)
        self.data["physicalData"] = ds.data["physicalData"]
        try:
            return self.data["physicalData"]["status"]
        except Exception:
            return None

    @property
    def download_path(self):
        """after ther physical datasets have been downloaded, this returns the relative path."""
        return self.__dict__.get("download_path", "")

    @property
    def _sftp_source_dir(self):
        """The SFTP directory is structured as follows:
        /SPACE/PROJECT/EXPERIMENT/permId

        For the current dataSet, this method returns the expected path
        """

        return os.path.join(self.experiment.identifier[1:], self.permId)

    def symlink(self, target_dir: str = None, replace_if_symlink_exists: bool = True):
        """replace_if_symlink_exists will replace the the target_dir
        in case it is an existing symlink
        Returns the absolute path of the symlink
        """

        if target_dir is None:
            target_dir = os.path.join(self.openbis.download_prefix, self.permId)

        target_dir_path = Path(target_dir)
        if target_dir_path.is_symlink() and replace_if_symlink_exists:
            target_dir_path.unlink()

        # create data/openbis-hostname
        os.makedirs(os.path.dirname(target_dir_path.absolute()), exist_ok=True)

        # make sure we got a mountpoint
        mountpoint_path = self.openbis.get_mountpoint()
        if mountpoint_path is None:
            try:
                mountpoint_path = self.openbis.mount()
            except ValueError as err:
                if "password" in str(err):
                    raise ValueError(
                        "openBIS instance cannot be mounted, no symlink possible"
                    )

        # construct the absolute path of our sftp source
        sftp_source_path = os.path.join(mountpoint_path, self._sftp_source_dir)

        # make sure our sftp source is really available
        # create symlink
        if os.path.exists(sftp_source_path):
            target_dir_path.symlink_to(sftp_source_path, target_is_directory=True)
            if VERBOSE:
                print(f"Symlink created: {target_dir} --> {sftp_source_path}")

            return str(target_dir_path.absolute())
        else:
            raise ValueError(
                f"Source path {sftp_source_path} does not exist, cannot create symlink"
            )

    @staticmethod
    def _file_set(target_dir: str) -> Set[str]:
        target_dir_path = Path(target_dir)
        return set(
            str(el.relative_to(target_dir_path))
            for el in target_dir_path.glob("**/*")
            if el.is_file()
        )

    def _is_symlink_or_physical(
            self,
            what: str,
            target_dir: str = None,
            expected_file_list: Optional[List[str]] = None,
    ):
        if target_dir is None:
            target_dir = os.path.join(self.openbis.download_prefix, self.permId)
        target_dir_path = Path(target_dir)

        target_file_set = self._file_set(target_dir)

        if expected_file_list is None:
            source_file_set = set(self.file_list)
        else:
            source_file_set = set(expected_file_list)

        res = source_file_set.issubset(target_file_set)
        if not res:
            return res
        elif what == "symlink":
            return target_dir_path.exists() and target_dir_path.is_symlink()
        elif what == "physical":
            return target_dir_path.exists() and not target_dir_path.is_symlink()
        else:
            raise ValueError("Unexpected error")

    is_symlink = partialmethod(
        _is_symlink_or_physical, what="symlink", expected_file_list=None
    )
    is_physical = partialmethod(_is_symlink_or_physical, what="physical")

    def archive(self, remove_from_data_store=True):
        fetchopts = {
            "removeFromDataStore": remove_from_data_store,
            "@type": "as.dto.dataset.archive.DataSetArchiveOptions",
        }
        self.archive_unarchive("archiveDataSets", fetchopts)
        if VERBOSE:
            print(f"DataSet {self.permId} archived")

    def unarchive(self):
        fetchopts = {"@type": "as.dto.dataset.unarchive.DataSetUnarchiveOptions"}
        self.archive_unarchive("unarchiveDataSets", fetchopts)
        if VERBOSE:
            print(f"DataSet {self.permId} unarchived")

    def archive_unarchive(self, method, fetchopts):
        payload = {}

        request = {
            "method": method,
            "params": [
                self.openbis.token,
                [{"permId": self.permId, "@type": "as.dto.dataset.id.DataSetPermId"}],
                dict(fetchopts),
            ],
        }
        resp = self.openbis._post_request(self._openbis.as_v3, request)
        return

    def set_properties(self, properties):
        """expects a dictionary of property names and their values.
        Does not save the dataset.
        """
        for prop in properties.keys():
            setattr(self.p, prop, properties[prop])

    set_props = set_properties

    def get_dataset_files(self, start_with=None, count=None, **properties):

        search_criteria = get_type_for_entity("dataSetFile", "search")
        search_criteria["operator"] = "AND"
        search_criteria["criteria"] = [
            {
                "criteria": [
                    {
                        "fieldName": "code",
                        "fieldType": "ATTRIBUTE",
                        "fieldValue": {
                            "value": self.permId,
                            "@type": "as.dto.common.search.StringEqualToValue",
                        },
                        "@type": "as.dto.common.search.CodeSearchCriteria",
                    }
                ],
                "operator": "OR",
                "@type": "as.dto.dataset.search.DataSetSearchCriteria",
            }
        ]

        fetchopts = get_fetchoption_for_entity("dataSetFile")

        request = {
            "method": "searchFiles",
            "params": [
                self.openbis.token,
                search_criteria,
                fetchopts,
            ],
        }
        full_url = urljoin(self._get_download_url(), DSS_ENDPOINT)
        resp = self.openbis._post_request_full_url(full_url, request)

        def create_data_frame(attrs, props, response):
            objects = response["objects"]
            parse_jackson(objects)

            attrs = [
                "dataSetPermId",
                "dataStore",
                "downloadUrl",
                "path",
                "directory",
                "fileLength",
                "checksumCRC32",
                "checksum",
                "checksumType",
            ]

            dataSetFiles = None
            if len(objects) == 0:
                dataSetFiles = DataFrame(columns=attrs)
            else:
                dataSetFiles = DataFrame(objects)
                dataSetFiles["downloadUrl"] = dataSetFiles["dataStore"].map(
                    extract_downloadUrl
                )
                dataSetFiles["checksumCRC32"] = (
                    dataSetFiles["checksumCRC32"]
                    .fillna(0.0)
                    .astype(int)
                    .map(signed_to_unsigned)
                )
                dataSetFiles["dataStore"] = dataSetFiles["dataStore"].map(extract_code)
                dataSetFiles["dataSetPermId"] = dataSetFiles["dataSetPermId"].map(
                    extract_permid
                )
            return dataSetFiles[attrs]

        return Things(
            openbis_obj=self.openbis,
            entity="dataSetFile",
            identifier_name="dataSetPermId",
            start_with=start_with,
            count=count,
            totalCount=resp.get("totalCount"),
            response=resp,
            df_initializer=create_data_frame,
        )

    def download(
            self,
            files=None,
            destination=None,
            create_default_folders=True,
            wait_until_finished=True,
            workers=10,
            linked_dataset_fileservice_url=None,
            content_copy_index=0
    ):
        """download the files of the dataSet.

        files -- a single file or a list of files. If no files are specified, all files of a given dataset are downloaded.
        destination -- if destination is specified, files are downloaded in __current_dir__/destination/permId/ If no destination is specified, the hostname is chosen instead of destination
        create_default_folders -- by default, this download method will automatically create destination/permId/original/DEFAULT. If create_default_folders is set to False, all these folders will be ommited. Use with care and by specifying the destination folder.
        workers -- Default: 10. Files are usually downloaded in parallel, using 10 workers by default.
        wait_unitl_finished -- True. If you want to immediately continue and run the download in background, set this to False.
        """

        if files == None:
            files = self.file_list
        elif isinstance(files, str):
            files = [files]

        if destination is None:
            destination = self.openbis.download_prefix
            # destination = self.openbis.hostname

        kind = None
        if "kind" in self.data:  # openBIS 18.6.x DTO
            kind = self.data["kind"]
        elif ("type" in self.data) and (
                "kind" in self.data["type"]
        ):  # openBIS 16.5.x DTO
            kind = self.data["type"]["kind"]

        if kind in ["PHYSICAL", "CONTAINER"]:
            if self.openbis.get_server_information().is_version_greater_than(3, 5):
                return self._download_fast_physical(files, destination, create_default_folders,
                                                    wait_until_finished)
            else:
                return self._download_physical(
                    files, destination, create_default_folders, wait_until_finished, workers
                )
        elif kind == "LINK":
            if linked_dataset_fileservice_url is None:
                raise ValueError(
                    "Can't download a LINK data set without the linked_dataset_fileservice_url parameters."
                )
            return self._download_link(
                files,
                destination,
                wait_until_finished,
                workers,
                linked_dataset_fileservice_url,
                content_copy_index,
            )
        else:
            raise ValueError(f"Can't download data set of kind {kind}.")

    def _download_fast_physical(
            self, files, destination, create_default_folders, wait_until_finished
    ):
        """Download for data sets of kind PHYSICAL using fast download scheme"""

        if create_default_folders:
            final_destination = os.path.join(destination, self.permId)
        else:
            final_destination = destination

        self.__dict__["download_path"] = final_destination

        download_url = self._get_download_url()

        fast_download = FastDownload(self.openbis.token, download_url, self.permId, files,
                                     final_destination, create_default_folders, wait_until_finished,
                                     self.openbis.verify_certificates, self.openbis.get_server_information(),
                                     wished_number_of_streams=4)
        return fast_download.download()

    def _download_physical(
            self, files, destination, create_default_folders, wait_until_finished, workers
    ):
        """Download for data sets of kind PHYSICAL."""

        final_destination = ""
        if create_default_folders:
            final_destination = os.path.join(destination, self.permId)
        else:
            final_destination = destination

        self.__dict__["download_path"] = final_destination

        download_url = self._get_download_url()
        base_url = download_url + "/datastore_server/" + self.permId + "/"
        with DataSetDownloadQueue(workers=workers) as queue:
            # get file list and start download
            for filename in files:
                fi_df = self.get_dataset_files().df
                file_size = fi_df[fi_df["path"] == filename]["fileLength"].values[0]
                download_url = base_url + filename + "?sessionID=" + self.openbis.token
                download_url = quote(download_url, safe=":/?=")
                filename_dest = ""
                if create_default_folders:
                    # create original/ or original/DEFAULT subfolders
                    filename_dest = os.path.join(final_destination, filename)
                else:
                    # ignore original/ and original/DEFAULT folders that come from openBIS
                    if filename.startswith("original/"):
                        filename = filename.replace("original/", "", 1)
                    if filename.startswith("DEFAULT/"):
                        filename = filename.replace("DEFAULT/", "", 1)
                    filename_dest = os.path.join(final_destination, filename)

                queue.put(
                    [
                        download_url,
                        filename,
                        filename_dest,
                        file_size,
                        self.openbis.verify_certificates,
                        "wb",
                    ]
                )

            # wait until all files have downloaded
            if wait_until_finished:
                queue.join()

            if VERBOSE:
                print(f"Files downloaded to: {os.path.join(final_destination)}")
            return final_destination

    def _download_link(
            self,
            files,
            destination,
            wait_until_finished,
            workers,
            linked_dataset_fileservice_url,
            content_copy_index,
    ):
        """Download for data sets of kind LINK.
        Requires the microservice server to be running at the given linked_dataset_fileservice_url.
        """

        with DataSetDownloadQueue(
                workers=workers, collect_files_with_wrong_length=True
        ) as queue:

            if content_copy_index >= len(self.data["linkedData"]["contentCopies"]):
                raise ValueError("Content Copy index out of range.")
            content_copy = self.data["linkedData"]["contentCopies"][content_copy_index]

            for filename in files:
                fi_df = self.get_dataset_files().df
                file_size = fi_df[fi_df["path"] == filename]["fileLength"].values[0]

                download_url = linked_dataset_fileservice_url
                download_url += "?sessionToken=" + self.openbis.token
                download_url += "&datasetPermId=" + self.data["permId"]["permId"]
                download_url += (
                        "&externalDMSCode=" + content_copy["externalDms"]["code"]
                )
                download_url += "&contentCopyPath=" + content_copy["path"].replace(
                    "/", "%2F"
                )
                download_url += "&datasetPathToFile=" + urllib.parse.quote(filename)

                filename_dest = os.path.join(destination, self.permId, filename)

                # continue download if file is not complete - do nothing if it is
                write_mode = "wb"
                if os.path.exists(filename_dest):
                    actual_size = os.path.getsize(filename_dest)
                    if actual_size == int(file_size):
                        continue
                    elif actual_size < int(file_size):
                        write_mode = "ab"
                        download_url += "&offset=" + str(actual_size)

                queue.put(
                    [
                        download_url,
                        filename,
                        filename_dest,
                        file_size,
                        self.openbis.verify_certificates,
                        write_mode,
                    ]
                )

            if wait_until_finished:
                queue.join()

            if VERBOSE:
                print(
                    "Files downloaded to: %s" % os.path.join(destination, self.permId)
                )
            return destination, queue.files_with_wrong_length

    @property
    def folder(self):
        return self.__dict__["folder"]

    @property
    def file_list(self):
        """Returns the list of files including their directories as an array of strings.
        Folders are not listed.
        """

        if self.is_new:
            return self.files
        else:
            fl = self.get_dataset_files().df
            return fl[fl["directory"] == False]["path"].to_list()

    @property
    def file_links(self):
        """Returns a dictionary of absolute file links for every file in this dataSet.
        As the link also contains a session token (sessionID), sharing this link might be
        a security risk. When the token is no longer valid, the link will no longer work either.
        """
        if self.is_new:
            return ""
        url = self.openbis.url
        location_part = self.physicalData.location.split("/")[-1]
        token = self.openbis.token

        file_links = {}
        for filepath in self.file_list:
            quoted_filepath = urllib.parse.quote(filepath, safe="")
            file_links[filepath] = (
                    "/".join([url, "datastore_server", location_part, quoted_filepath])
                    + "?sessionID="
                    + token
            )

        return file_links

    @property
    def rel_file_links(self):
        """Returns a dictionary of relative file links for every file in this dataSet. These relative file link can be embedded in a <img src="{rel_link}">
        element within a XML property. If the dataSet file happens to be a picture, in ELN-LIMS, the picture will be displayed inline.
        """
        if self.is_new:
            return ""
        url = self.openbis.url
        location_part = self.physicalData.location.split("/")[-1]

        rel_file_links = {}
        for filepath in self.file_list:
            quoted_filepath = urllib.parse.quote(filepath, safe="")
            rel_file_links[filepath] = "/".join(
                ["/datastore_server", location_part, quoted_filepath]
            )

        return rel_file_links

    def get_files(self, start_folder="/"):
        """Returns a DataFrame of all files in this dataset"""
        if start_folder.startswith("/"):
            start_folder = start_folder[1:]
        file_list = self.get_dataset_files().df
        file_list[file_list["path"].str.startswith(start_folder)]
        new_file_list = file_list[
            ["directory", "path", "fileLength", "checksumCRC32"]
        ].rename(
            columns={
                "directory": "isDirectory",
                "path": "pathInDataSet",
                "fileLength": "fileSize",
                "checksumCRC32": "crc32Checksum",
            }
        )
        return new_file_list

    def _get_download_url(self):
        download_url = ""
        if "downloadUrl" in self.data["dataStore"]:
            download_url = self.data["dataStore"]["downloadUrl"]
        else:
            # fallback, if there is no dataStore defined
            datastores = self.openbis.get_datastores()
            download_url = datastores["downloadUrl"][0]
        return download_url

    def get_file_list(self, recursive=True, start_folder="/"):
        """Lists all files of a given dataset. You can specifiy a start_folder other than "/".
        By default, all directories and their containing files are listed recursively. You can
        turn off this option by setting recursive=False.
        """
        print("This method is deprecated. Consider using get_files() instead")
        request = {
            "method": "listFilesForDataSet",
            "params": [
                self.openbis.token,
                self.permId,
                start_folder,
                recursive,
            ],
            "id": "1",
        }
        download_url = self._get_download_url()
        resp = requests.post(
            download_url + "/datastore_server/rmi-dss-api-v1.json",
            json.dumps(request),
            verify=self.openbis.verify_certificates,
        )

        if resp.ok:
            data = resp.json()
            if "error" in data:
                raise ValueError("Error from openBIS: " + data["error"]["message"])
            elif "result" in data:
                return data["result"]
            else:
                raise ValueError(
                    "request to openBIS did not return either result nor error"
                )
        else:
            raise ValueError("internal error while performing post request")

    def _generate_plugin_request(self, dss, permId=None):
        """generates a request to activate the dataset-uploader ingestion plugin to
        register our files as a new dataset
        """

        sample_identifier = None
        if self.sample is not None:
            sample_identifier = self.sample.identifier

        experiment_identifier = None
        if self.experiment is not None:
            experiment_identifier = self.experiment.identifier

        parentIds = self.parents
        if parentIds is None:
            parentIds = []

        dataset_type = self.type.code
        properties = self.formatter.format(self.props.all_nonempty())

        request = {
            "method": "createReportFromAggregationService",
            "params": [
                self.openbis.token,
                dss,
                PYBIS_PLUGIN,
                {
                    "permId": permId,
                    "method": "insertDataSet",
                    "sampleIdentifier": sample_identifier,
                    "experimentIdentifier": experiment_identifier,
                    "dataSetType": dataset_type,
                    "folderName": self.folder,
                    "fileNames": self.files_in_wsp,
                    "isZipDirectoryUpload": self.isZipDirectoryUpload,
                    "properties": properties,
                    "parentIdentifiers": parentIds,
                },
            ],
        }
        return request

    def save(self, permId=None):
        for prop_name, prop in self.props._property_names.items():
            if prop["mandatory"]:
                if (
                        getattr(self.props, prop_name) is None
                        or getattr(self.props, prop_name) == ""
                ):
                    raise ValueError(
                        f"Property '{prop_name}' is mandatory and must not be None"
                    )

        if self.is_new:
            data_stores = self.openbis.get_datastores()

            if self.sample is None and self.experiment is None:
                raise ValueError(
                    "A DataSet must be either connected to a Sample or an Experiment"
                )

            if self.kind == "PHYSICAL":
                if self.files is None or len(self.files) == 0:
                    raise ValueError(
                        "Cannot register a dataset without a file. Please provide at least one file"
                    )
                if self.openbis.get_server_information().is_version_greater_than(3, 5):
                    return self._upload_v3(data_stores)

                return self._upload_v1(permId, data_stores)
            # CONTAINER
            else:
                if self.files is not None and len(self.files) > 0:
                    raise ValueError(
                        "DataSets of kind CONTAINER or LINK cannot contain data"
                    )

                request = self._new_attrs()

                # if no code for the container was provided, let openBIS
                # generate the code automatically
                if self.code is None or self.code == "":
                    request["params"][1][0]["autoGeneratedCode"] = True
                else:
                    request["params"][1][0]["autoGeneratedCode"] = False

                props = self.formatter.format(self.p._all_props())
                DSpermId = data_stores["code"][0]
                request["params"][1][0]["properties"] = props
                request["params"][1][0]["dataStoreId"] = {
                    "permId": DSpermId,
                    "@type": "as.dto.datastore.id.DataStorePermId",
                }

                version = self.openbis.get_server_information().openbis_version
                if version is not None:
                    if 'SNAPSHOT' not in version and not version.startswith(
                            '6') and 'UNKNOWN' not in version:
                        if (request['method'] in ('createDataSetTypes', 'createDataSets')
                                and 'metaData' in request['params'][1][0]):
                            del request['params'][1][0]['metaData']


                resp = self.openbis._post_request(self.openbis.as_v3, request)

                if VERBOSE:
                    print("DataSet successfully created.")
                new_dataset_data = self.openbis.get_dataset(
                    resp[0]["permId"], only_data=True
                )
                self._set_data(new_dataset_data)
                return self

        # updating the DataSET
        else:
            request = self._up_attrs()
            props = self.formatter.format(self.p._all_props())
            request["params"][1][0]["properties"] = props

            version = self.openbis.get_server_information().openbis_version
            if version is not None:
                if 'SNAPSHOT' not in version and not version.startswith(
                        '6') and 'UNKNOWN' not in version:
                    if (request['method'] in ('updateDataSetTypes', 'updateDataSets')
                            and 'metaData' in request['params'][1][0]):
                        del request['params'][1][0]['metaData']

            self.openbis._post_request(self.openbis.as_v3, request)
            if VERBOSE:
                print("DataSet successfully updated.")

    def _upload_v1(self, permId, datastores):
        # for uploading phyiscal data, we first upload it to the session workspace
        self.upload_files_v1(
            datastore_url=datastores["downloadUrl"][0],
            files=self.files,
            folder="",
            wait_until_finished=True,
        )

        # activate the ingestion plugin, as soon as the data is uploaded
        # this will actually register the dataset in the datastore and the AS
        request = self._generate_plugin_request(
            dss=datastores["code"][0],
            permId=permId,
        )
        resp = self.openbis._post_request(self.openbis.reg_v1, request)
        if resp["rows"][0][0]["value"] == "OK":
            permId = resp["rows"][0][2]["value"]
            if permId is None or permId == "":
                self.__dict__["is_new"] = False
                if VERBOSE:
                    print(
                        "DataSet successfully created. Because you connected to an openBIS version older than 16.05.04, you cannot update the object."
                    )
            else:
                new_dataset_data = self.openbis.get_dataset(
                    permId, only_data=True
                )
                self._set_data(new_dataset_data)
                if VERBOSE:
                    print("DataSet successfully created.")
                return self
        else:
            print(json.dumps(request))
            raise ValueError(
                "Error while creating the DataSet: "
                + resp["rows"][0][1]["value"]
            )

    def _upload_v3(self, data_stores):
        datastore_url = data_stores["downloadUrl"][0]
        # for uploading phyiscal data, we first upload it to the session workspace
        upload_id = self.upload_files_v3(
            datastore_url=datastore_url,
            files=self.files,
            wait_until_finished=True,
        )
        type_code = self.type.code
        if type_code.startswith('$'):
            type_code = type_code[1:]

        props = self.formatter.format(self.props.all_nonempty())
        param = {
            "@type": "dss.dto.dataset.create.UploadedDataSetCreation",
            "@id": "1",
            "typeId": {
                "@type": "as.dto.entitytype.id.EntityTypePermId",
                "@id": "2",
                "permId": type_code,
                "entityKind": "DATA_SET"},

            "properties": props,
            "parentIds": [],
            "uploadId": upload_id
        }

        if self.experiment is not None:
            param["experimentId"] = {
                "@type": "as.dto.experiment.id.ExperimentIdentifier",
                "@id": "3",
                "identifier": self.experiment.identifier
            }
        if self.sample is not None:
            param["sampleId"] = {
                "@type": "as.dto.sample.id.SamplePermId",
                "@id": "4",
                "permId": self.sample.permId
            }
        # TODO: check if this part is needed
        parent_ids = self.parents
        if parent_ids is None:
            parent_ids = []
        counter = 5
        for parent_id in parent_ids:
            param["parentIds"] += [{
                "@type": "as.dto.dataset.id.DataSetPermId",
                "@id": str(counter),
                "permId": parent_id
            }]
            counter += 1

        request = {
            "method": "createUploadedDataSet",
            "params": [self.openbis.token, param]
        }

        resp = self.openbis._post_request_full_url(urljoin(datastore_url, self.openbis.dss_v3),
                                                   request)
        if "permId" in resp:
            permId = resp["permId"]
            if permId is None or permId == "":
                self.__dict__["is_new"] = False
                if VERBOSE:
                    print(
                        "DataSet successfully created. Because you connected to an openBIS version older than 16.05.04, you cannot update the object."
                    )
            else:
                new_dataset_data = self.openbis.get_dataset(
                    permId, only_data=True
                )
                self._set_data(new_dataset_data)
                if VERBOSE:
                    print("DataSet successfully created.")
                return self
        else:
            print(json.dumps(request))
            raise ValueError(
                "Error while creating the DataSet: "
                + resp["rows"][0][1]["value"]
            )

    def zipit(self, file_or_folder, zipf):
        """Takes a directory or a file, and a zipfile instance. For every file that is encountered,
        we issue the write() method to add that file to the zipfile.
        If we have a directory, we walk that directory and add every file inside it,
        including the starting folder name.
        """
        if os.path.isfile(file_or_folder):
            # if a file is provided, we want to always store it in the root of the zip file
            # ../../somedir/file.txt       -->   file.txt
            (realpath, filename) = os.path.split(os.path.realpath(file_or_folder))
            zipf.write(file_or_folder, filename)
        elif os.path.isdir(file_or_folder):
            # if a directory is provided, we want to store it (and its content) also in the root of the zip file
            # ../../somedir/               -->   somedir/
            # ../../somedir/other/file.txt -->   somedir/other/file.txt
            (head, tail) = os.path.split(os.path.realpath(file_or_folder))
            for dirpath, dirnames, filenames in os.walk(file_or_folder):
                realpath = os.path.realpath(dirpath)
                for filename in filenames:
                    zipf.write(
                        os.path.relpath(
                            os.path.join(dirpath, filename),
                            os.path.join(filename, ".."),
                        ),
                        os.path.join(realpath[len(head) + 1:], filename),
                    )

    def upload_files_v1(
            self, datastore_url=None, files=None, folder=None, wait_until_finished=False
    ):

        if datastore_url is None:
            datastore_url = self.openbis._get_dss_url()
        if files is None:
            raise ValueError("Please provide a filename.")

        if folder is None:
            # create a unique foldername
            folder = time.strftime("%Y-%m-%d_%H-%M-%S")

        if isinstance(files, str):
            files = [files]

        contains_dir = False
        for f in files:
            if os.path.isdir(f):
                contains_dir = True

        if contains_dir:
            # if the file list contains at least one directory, we need to zip the
            # whole thing in order to get it safely to openBIS.
            file_ending = "".join(
                random.choice(
                    "0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"
                )
                for i in range(6)
            )
            filename = time.strftime("%Y-%m-%d_%H-%M-%S") + file_ending + ".zip"
            buf = ZipBuffer(
                openbis_obj=self.openbis, host=datastore_url, filename=filename
            )
            zipf = zipfile.ZipFile(file=buf, mode="w", compression=zipfile.ZIP_DEFLATED)
            for file_or_folder in files:
                self.zipit(file_or_folder, zipf)
            # self.__dict__['folder'] = '/'
            self.__dict__["files_in_wsp"] = [filename]
            self.__dict__["isZipDirectoryUpload"] = True
            return self.files_in_wsp

        # define a queue to handle the upload threads
        with DataSetUploadQueue() as queue:

            real_files = []
            for filename in files:
                if os.path.isdir(filename):
                    real_files.extend(
                        [
                            os.path.join(dp, f)
                            for dp, dn, fn in os.walk(os.path.expanduser(filename))
                            for f in fn
                        ]
                    )
                else:
                    real_files.append(os.path.join(filename))

            # compose the upload-URL and put URL and filename in the upload queue
            for filename in real_files:
                file_in_wsp = os.path.join(folder, os.path.basename(filename))
                url_filename = os.path.join(
                    folder, urllib.parse.quote(os.path.basename(filename))
                )
                self.files_in_wsp.append(file_in_wsp)

                upload_url = (
                        datastore_url
                        + "/datastore_server/session_workspace_file_upload"
                        + "?filename="
                        + url_filename
                        + "&id=1"
                        + "&startByte=0&endByte=0"
                        + "&sessionID="
                        + self.openbis.token
                )
                queue.put([upload_url, filename, self.openbis.verify_certificates])

            # wait until all files have uploaded
            if wait_until_finished:
                queue.join()

            # return files with full path in session workspace
            return self.files_in_wsp

    def upload_files_v3(
            self, files, datastore_url=None, folder=None, wait_until_finished=False
    ):
        if datastore_url is None:
            datastore_url = self.openbis._get_dss_url()
        if files is None:
            raise ValueError("Please provide a filename.")

        if isinstance(files, str):
            files = [files]

        upload_id = str(uuid.uuid4())
        if len(files) == 1:
            if folder is None:
                folder = upload_id
            else:
                folder = os.path.join(upload_id, folder)

        else:
            if folder is None:
                folder = os.path.join(upload_id, 'default')
            else:
                folder = os.path.join(upload_id, folder)

        if len(files) == 0:
            raise ValueError("Please provide a filename.")

        # define a queue to handle the upload threads
        with DataSetUploadQueueNew(datastore_url) as queue:
            real_files = []
            for filename in files:
                if os.path.isdir(filename):
                    pardir = os.path.join(filename, os.pardir)
                    for root, dirs, files in os.walk(os.path.expanduser(filename)):
                        path = os.path.relpath(root, pardir)
                        for file in files:
                            real_files.append((path, os.path.join(root, file)))
                        if not files:
                            # append empty folder
                            real_files.append((path, ""))
                else:
                    real_files.append(("", os.path.join(filename)))

            # compose the upload-URL and put URL and filename in the upload queue
            for filename in real_files:
                file_in_wsp = os.path.join(folder, filename[0], os.path.basename(filename[1]))
                url_filename = os.path.join(
                    folder, filename[0], urllib.parse.quote(os.path.basename(filename[1]))
                )
                # Windows directory separator is \
                url_filename = '/'.join(url_filename.split('\\'))
                self.files_in_wsp.append(file_in_wsp)

                is_empty_folder = filename[1] == ''
                if is_empty_folder:
                    upload_url = (f'{datastore_url}{SESSION_WORKSPACE}'
                                  f'?filename={url_filename}'
                                  f'&id={1}'
                                  f'&startByte={0}&endByte={0}'
                                  f'&emptyFolder={True}'
                                  f'&sessionID={self.openbis.token}')
                    queue.put([upload_url, filename, self.openbis.verify_certificates, True, False,
                               []])
                else:
                    file_size = os.path.getsize(filename[1])
                    count = 1
                    size = 1024 * 1024 * 10  # 10MB
                    if file_size > size:
                        for i in range(0, file_size, size):
                            start_byte = i
                            end_byte = min(i + size - 1, file_size)
                            upload_url = (f'{datastore_url}{SESSION_WORKSPACE}'
                                          f'?filename={url_filename}'
                                          f'&id={count}'
                                          f'&startByte={start_byte}&endByte={end_byte}'
                                          f'&emptyFolder={False}'
                                          f'&sessionID={self.openbis.token}')
                            queue.put(
                                [upload_url, filename, self.openbis.verify_certificates, False,
                                 True,
                                 [start_byte, end_byte]])
                            count += 1
                    else:
                        upload_url = (
                                datastore_url
                                + "/datastore_server/session_workspace_file_upload"
                                + "?filename="
                                + url_filename
                                + "&id="
                                + str(count)
                                + "&startByte=0&endByte=" + str(file_size)
                                + "&emptyFolder=False"
                                + "&sessionID="
                                + self.openbis.token
                        )
                        queue.put(
                            [upload_url, filename, self.openbis.verify_certificates, False, False,
                             []])

            # wait until all files have uploaded
            if wait_until_finished:
                try:
                    queue.join()
                except BaseException as e:
                    raise e

            # return files with full path in session workspace
            return upload_id


class PropagatingThread(Thread):
    def run(self):
        self.exc = None
        try:
            self.ret = self._target(*self._args, **self._kwargs)
        except BaseException as e:
            self.exc = e

    def join(self, timeout=None):
        super(PropagatingThread, self).join(timeout)
        if self.exc:
            raise self.exc
        return self.ret


class DataSetUploadQueueNew:
    """Structure for uploading files to OpenBIS in separate threads.
    It works as a queue where each item is a single file upload. It allows to upload files using v1
    and v3 api. V3 api uses multipart schema for file upload, whereas V1 api makes sue of the body"""

    def create_session(self, url_base):
        """Create a session object to handle retries in case of server failure"""
        session = requests.Session()
        retries = Retry(total=REQUEST_RETRIES_COUNT, backoff_factor=1,
                        status_forcelist=[502, 503, 504])
        session.mount(url_base, HTTPAdapter(max_retries=retries))
        return session

    def __init__(self, url_base, workers=10):
        # maximum files to be uploaded at once
        self.upload_queue = Queue()
        self.workers = workers
        self.session = self.create_session(url_base)
        self.threads = []
        # define number of threads and start them
        for t in range(workers):
            t = PropagatingThread(target=self.upload_file)
            self.threads += [t]
            t.start()

    def __enter__(self, *args, **kwargs):
        return self

    def __exit__(self, *args, **kwargs):
        """This method is called at the end of a with statement."""
        # stop the workers
        for i in range(self.workers):
            self.upload_queue.put(None)

    def put(self, things):
        """expects a list [url, filename] which is put into the upload queue"""
        self.upload_queue.put(things)

    def join(self):
        """needs to be called if you want to wait for all uploads to be finished"""
        # block until all tasks are done
        self.upload_queue.join()
        for t in self.threads:
            if t.exc is not None:
                raise t.exc

    def upload_file(self):
        while True:
            # get the next item in the queue
            queue_item = self.upload_queue.get()
            if queue_item is None:
                # when we call the .join() method of the DataSetUploadQueue and empty the queue
                break
            upload_url, filename, verify_certificates, is_empty_folder, partial, bytes_range = queue_item

            try:
                # upload the file to our DSS session workspace
                if is_empty_folder:
                    resp = self.session.post(upload_url, verify=verify_certificates)
                    resp.raise_for_status()
                else:
                    if partial:
                        with open(filename[1], "rb") as f:
                            f.seek(bytes_range[0])
                            data = f.read(bytes_range[1] - bytes_range[0] + 1)
                            resp = self.session.post(upload_url, data=data,
                                                     verify=verify_certificates)
                            resp.raise_for_status()
                    else:
                        file_size = os.path.getsize(filename[1])
                        with open(filename[1], "rb") as f:
                            resp = self.session.post(upload_url, data=f, verify=verify_certificates)
                            resp.raise_for_status()
                            data = resp.json()
                            if file_size != int(data["size"]):
                                raise ValueError(
                                    f'size of file uploaded: {file_size} != data received: {int(data["size"])}'
                                )
            except ValueError as e:
                with self.upload_queue.mutex:
                    self.upload_queue.all_tasks_done.notify_all()
                raise e
            finally:
                # Tell the queue that we are done
                self.upload_queue.task_done()
        return True


class DataSetUploadQueue:
    """Structure for uploading files to OpenBIS in separate threads.
    It works as a queue where each item is a single file upload. It allows to upload files using v1
    and v3 api. V3 api uses multipart schema for file upload, whereas V1 api makes sue of the body"""

    def __init__(self, workers=20, multipart=False):
        # maximum files to be uploaded at once
        self.upload_queue = Queue()
        self.workers = workers
        self.multipart = multipart

        # define number of threads and start them
        for t in range(workers):
            t = Thread(target=self.upload_file)
            t.start()

    def __enter__(self, *args, **kwargs):
        return self

    def __exit__(self, *args, **kwargs):
        """This method is called at the end of a with statement."""
        # stop the workers
        for i in range(self.workers):
            self.upload_queue.put(None)

    def put(self, things):
        """expects a list [url, filename] which is put into the upload queue"""
        self.upload_queue.put(things)

    def join(self):
        """needs to be called if you want to wait for all uploads to be finished"""
        # block until all tasks are done
        self.upload_queue.join()

    def upload_file(self):
        while True:
            # get the next item in the queue
            queue_item = self.upload_queue.get()
            if queue_item is None:
                # when we call the .join() method of the DataSetUploadQueue and empty the queue
                break
            upload_url, filename, verify_certificates = queue_item

            file_size = os.path.getsize(filename)

            if self.multipart is True:
                file = {filename: open(filename, "rb")}
                resp = requests.post(upload_url, files=file, verify=verify_certificates)
                resp.raise_for_status()
            else:
                # upload the file to our DSS session workspace
                with open(filename, "rb") as f:
                    resp = requests.post(upload_url, data=f, verify=verify_certificates)
                    resp.raise_for_status()
                    data = resp.json()
                    if file_size != int(data["size"]):
                        raise ValueError(
                            f'size of file uploaded: {file_size} != data received: {int(data["size"])}'
                        )

            # Tell the queue that we are done
            self.upload_queue.task_done()


class ZipBuffer(object):
    """A file-like object for zipfile.ZipFile to write into.
    zipfile invokes the write method to store its zipped content.
    We will send this content directly to the session_workspace as a POST request.
    Used by V1 API only.
    """

    def __init__(self, openbis_obj, host, filename):
        self.openbis = openbis_obj
        self.startByte = 0
        self.endByte = 0
        self.filename = filename
        self.upload_url = (
                host + "/datastore_server/session_workspace_file_upload?"
                       "filename={}"
                       "&id=1"
                       "&startByte={}"
                       "&endByte={}"
                       "&sessionID={}"
        )
        self.session = Session()

    def write(self, data):

        self.startByte = self.endByte
        self.endByte += len(data)
        attempts = 0

        while True:
            attempts += 1
            resp = self.session.post(
                url=self.upload_url.format(
                    self.filename, self.startByte, self.endByte, self.openbis.token
                ),
                data=data,
                verify=self.openbis.verify_certificates,
            )
            if resp.status_code == 200:
                break
            if attempts > 10:
                raise Exception("Upload failed after more than 10 attempts")

    def tell(self):
        """Return the current stream position."""
        return self.endByte

    def flush(self):
        """Flush the write buffers of the stream if applicable."""
        self.session.close()
        pass


class DataSetDownloadQueue:
    """Special queue structure for multithreaded downloading files using V1 API."""

    def __init__(self, workers=20, collect_files_with_wrong_length=False):
        self.collect_files_with_wrong_length = collect_files_with_wrong_length
        # maximum files to be downloaded at once
        self.workers = workers
        self.download_queue = Queue()
        self.files_with_wrong_length = []

        # define number of threads
        for i in range(workers):
            thread = Thread(target=self.download_file)
            thread.start()

    def __enter__(self, *args, **kwargs):
        return self

    def __exit__(self, *args, **kwargs):
        """This method is called at the end of a with statement."""
        # stop all workers
        for i in range(self.workers):
            self.download_queue.put(None)

    def put(self, things):
        """expects a list [url, filename] which is put into the download queue"""
        self.download_queue.put(things)

    def join(self):
        """needs to be called if you want to wait for all downloads to be finished"""
        self.download_queue.join()

    def download_file(self):
        while True:
            try:
                queue_item = self.download_queue.get()
                if queue_item is None:
                    # when we call the .join() method of the DataSetDownloadQueue and empty the queue
                    break
                (
                    url,
                    filename,
                    filename_dest,
                    file_size,
                    verify_certificates,
                    write_mode,
                ) = queue_item
                # create the necessary directory structure if they don't exist yet
                os.makedirs(os.path.dirname(filename_dest), exist_ok=True)

                # request the file in streaming mode
                r = requests.get(url, stream=True, verify=verify_certificates)
                if r.ok == False:
                    raise ValueError(
                        f"Could not download from {url}: HTTP {r.status_code}. Reason: {r.reason}"
                    )

                with open(filename_dest, write_mode) as fh:
                    for chunk in r.iter_content(chunk_size=1024 * 1024):
                        if chunk:  # filter out keep-alive new chunks
                            fh.write(chunk)

                r.raise_for_status()
                actual_file_size = os.path.getsize(filename_dest)
                if actual_file_size != int(file_size):
                    if self.collect_files_with_wrong_length:
                        self.files_with_wrong_length.append(filename)
                    else:
                        print(
                            f"WARNING! File {filename_dest} has the wrong length: Expected: {int(file_size)} Actual size: {actual_file_size}"
                        )
                        print(
                            "REASON: The connection has been silently dropped upstreams.",
                            "Please check the http timeout settings of the openBIS datastore server",
                        )
            except Exception as err:
                print(f"ERROR while writing file {filename_dest}: {err}")

            finally:
                self.download_queue.task_done()


class PhysicalData:
    def __init__(self, data=None):
        if data is None:
            data = {}
        self.data = data
        self.attrs = [
            "speedHint",
            "complete",
            "shareId",
            "size",
            "fileFormatType",
            "storageFormat",
            "location",
            "presentInArchive",
            "storageConfirmation",
            "locatorType",
            "status",
        ]

    def __dir__(self):
        return self.attrs

    def __getattr__(self, name):
        if name in self.attrs:
            if name in self.data:
                return self.data[name]

    def __getitem__(self, key):
        if key in self.attrs:
            if key in self.data:
                return self.data[key]

    def _repr_html_(self):
        html = """
            <table border="1" class="dataframe">
            <thead>
                <tr style="text-align: right;">
                <th>attribute</th>
                <th>value</th>
                </tr>
            </thead>
            <tbody>
        """

        for attr in self.attrs:
            html += f"<tr> <td>{attr}</td> <td>{getattr(self, attr, '')}</td> </tr>"

        html += """
            </tbody>
            </table>
        """
        return html

    def __repr__(self):
        headers = ["attribute", "value"]
        lines = []
        for attr in self.attrs:
            lines.append([attr, getattr(self, attr, "")])
        return tabulate(lines, headers=headers)


class LinkedData:
    def __init__(self, data=None):
        self.data = data if data is not None else []
        self.attrs = ["externalCode", "contentCopies"]

    def __dir__(self):
        return self.attrs

    def __getattr__(self, name):
        if name in self.attrs:
            if name in self.data:
                return self.data[name]
        else:
            return ""



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#   Copyright ETH 2023 - 2024 Zürich, Scientific IT Services
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#

"""
imaging.py

Work with Openbis Imaging technology with Python.

"""
import abc
import json
import base64
import threading



class AtomicIncrementer:
    def __init__(self, value=0):
        self._value = int(value)
        self._lock = threading.Lock()

    def inc(self, d=1):
        with self._lock:
            self._value += int(d)
            return self._value


class AbstractImagingClass(metaclass=abc.ABCMeta):
    def to_json(self):

        c = AtomicIncrementer()

        def dictionary_creator(x):
            dictionary = x.__dict__
            val = c.inc()
            dictionary['@id'] = val
            return dictionary

        return json.dumps(self, default=dictionary_creator, sort_keys=True, indent=4)

    def __str__(self):
        return json.dumps(self.__dict__, default=lambda x: x.__dict__)

    def __repr__(self):
        return json.dumps(self.__dict__, default=lambda x: x.__dict__)


class AbstractImagingRequest(AbstractImagingClass, metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def _validate_data(self):
        return

class ImagingSemanticAnnotation(AbstractImagingClass):
    ontologyId: str
    ontologyVersion: str
    ontologyAnnotationId: str

    def __init__(self, ontologyId=None, ontologyVersion=None, ontologyAnnotationId=None):
        self.__dict__["@type"] = "imaging.dto.ImagingSemanticAnnotation"
        self.ontologyId = ontologyId
        self.ontologyVersion = ontologyVersion
        self.ontologyAnnotationId = ontologyAnnotationId

    @classmethod
    def from_dict(cls, data):
        if data is None:
            return None
        if "@id" in data:
            del data["@id"]
        semantic_annotation = cls(None, None, None)
        for prop in cls.__annotations__.keys():
            attribute = data.get(prop)
            semantic_annotation.__dict__[prop] = attribute
        return semantic_annotation

class ImagingDataSetFilter(AbstractImagingClass):
    name: str
    parameters: dict

    def __init__(self, name, parameters=None):
        self.__dict__["@type"] = "imaging.dto.ImagingDataSetFilter"
        self.name = name
        self.parameters = parameters if parameters is not None else dict()

    @classmethod
    def from_dict(cls, data):
        if data is None:
            return None
        if "@id" in data:
            del data["@id"]
        imaging_filter = cls('', None)
        for prop in cls.__annotations__.keys():
            attribute = data.get(prop)
            imaging_filter.__dict__[prop] = attribute
        return imaging_filter

class ImagingDataSetPreview(AbstractImagingRequest):
    config: dict
    format: str
    bytes: str
    width: int
    height: int
    index: int
    show: bool
    metadata: dict
    comment: str
    tags: list
    filterConfig: list

    def __init__(self, preview_format, config=None, metadata=None, index=0, comment="", tags=[], filterConfig=[]):
        self.__dict__["@type"] = "imaging.dto.ImagingDataSetPreview"
        self.bytes = None
        self.format = preview_format
        self.config = config if config is not None else dict()
        self.metadata = metadata if metadata is not None else dict()
        self.index = index
        self.comment = comment
        self.tags = tags if tags is not None else []
        self.filterConfig = filterConfig if filterConfig is not None else []
        self._validate_data()

    def set_preview_image_bytes(self, width, height, bytes):
        self.width = width
        self.height = height
        self.bytes = bytes

    def _validate_data(self):
        assert self.format is not None, "Format can not be null"

    def save_to_file(self, file_path):
        assert self.bytes is not None, "There is no image information!"
        img_data = bytearray(self.bytes, encoding='utf-8')
        with open(file_path, "wb") as fh:
            fh.write(base64.decodebytes(img_data))

    @classmethod
    def from_dict(cls, data):
        if data is None:
            return None
        if "@id" in data:
            del data["@id"]
        preview = cls('', None, None)
        for prop in cls.__annotations__.keys():
            attribute = data.get(prop)
            preview.__dict__[prop] = attribute
        return preview

class ImagingDataSetExportConfig(AbstractImagingClass):
    archiveFormat: str
    imageFormat: str
    resolution: str
    include: list

    def __init__(self, archive_format, image_format, resolution, include=None):
        if include is None:
            include = ["IMAGE", "RAW_DATA"]
        self.__dict__["@type"] = "imaging.dto.ImagingDataSetExportConfig"
        self.imageFormat = image_format
        self.archiveFormat = archive_format
        if resolution is None:
            resolution = "original"
        self.resolution = resolution
        self.include = include
        self._validate_data()

    def _validate_data(self):
        assert self.imageFormat is not None, "image format can not be null"
        assert self.archiveFormat is not None, "image format can not be null"

    @classmethod
    def from_dict(cls, data):
        if data is None:
            return None
        if "@id" in data:
            del data["@id"]
        preview = cls(None, None, None)
        for prop in cls.__annotations__.keys():
            attribute = data.get(prop)
            preview.__dict__[prop] = attribute
        return preview



class ImagingDataSetExport(AbstractImagingRequest):
    config: ImagingDataSetExportConfig
    metadata: dict

    def __init__(self, config, metadata=None):
        self.__dict__["@type"] = "imaging.dto.ImagingDataSetExport"
        self.config = config
        self.metadata = metadata if metadata is not None else dict()
        self._validate_data()

    def _validate_data(self):
        assert self.config is not None, "Config can not be null"



class ImagingDataSetMultiExport(AbstractImagingRequest):
    permId: str
    imageIndex: int
    previewIndex: int
    config: ImagingDataSetExportConfig
    metadata: dict

    def __init__(self, permId, imageIndex, previewIndex, config, metadata=None):
        self.__dict__["@type"] = "imaging.dto.ImagingDataSetMultiExport"
        self.permId = permId
        self.imageIndex = imageIndex
        self.previewIndex = previewIndex
        self.config = config
        self.metadata = metadata if metadata is not None else dict()
        self._validate_data()

    def _validate_data(self):
        assert self.permId is not None, "PermId can not be null"
        assert self.imageIndex is not None, "imageIndex can not be null"
        assert self.previewIndex is not None, "previewIndex can not be null"


class ImagingDataSetControlVisibility(AbstractImagingClass):
    label: str
    values: list
    range: list
    unit: str

    def __init__(self, label: str, values: list, values_range: list, unit: str = None):
        self.__dict__["@type"] = "imaging.dto.ImagingDataSetControlVisibility"
        self.label = label
        self.values = values
        self.range = values_range
        self.unit = unit

    @classmethod
    def from_dict(cls, data):
        if data is None:
            return None
        if "@id" in data:
            del data["@id"]
        control = cls(None, None, None, None)
        for prop in cls.__annotations__.keys():
            attribute = data.get(prop)
            control.__dict__[prop] = attribute
        return control


class ImagingDataSetControl(AbstractImagingClass):
    label: str
    section: str
    type: str
    values: list
    unit: str
    range: list
    multiselect: bool
    playable: bool
    speeds: list
    visibility: list
    metadata: dict
    semanticAnnotation: ImagingSemanticAnnotation

    def __init__(self, label: str, control_type: str, section: str = None, values: list = None,
                 unit: str = None, values_range: list = None, multiselect: bool = None,
                 playable: bool = False, speeds: list = None,
                 visibility: list = None, metadata: dict = None, semanticAnnotation: ImagingSemanticAnnotation = None):
        self.__dict__["@type"] = "imaging.dto.ImagingDataSetControl"
        self.label = label
        self.type = control_type
        self.section = section
        self.unit = unit
        if control_type.lower() in ["slider", "range"]:
            self.range = values_range
        elif control_type.lower() in ["dropdown", "colormap"]:
            self.values = values
            if multiselect is None:
                self.multiselect = False
            else:
                self.multiselect = multiselect

        if playable is True:
            self.playable = True
            self.speeds = speeds
        self.visibility = visibility
        self.metadata = metadata
        self.semanticAnnotation = semanticAnnotation

    @classmethod
    def from_dict(cls, data):
        if data is None:
            return None
        if "@id" in data:
            del data["@id"]
        control = cls(None, "", None, None)
        for prop in cls.__annotations__.keys():
            attribute = data.get(prop)
            if attribute is not None:
                if prop == 'visibility':
                    attribute = [ImagingDataSetControlVisibility.from_dict(visibility) for visibility in attribute]
                elif prop == 'semanticAnnotation':
                    attribute = ImagingSemanticAnnotation.from_dict(attribute)
            control.__dict__[prop] = attribute
        return control


class ImagingDataSetConfig(AbstractImagingClass):
    adaptor: str
    version: float
    speeds: list
    resolutions: list
    playable: bool
    exports: list
    inputs: list
    metadata: dict
    filters: dict
    filterSemanticAnnotation: dict

    def __init__(self, adaptor: str, version: float, resolutions: list, playable: bool,
                 speeds: list = None, exports: list = None,
                 inputs: list = None, metadata: dict = None, filters: dict = None,
                 filterSemanticAnnotation: dict = None):
        self.__dict__["@type"] = "imaging.dto.ImagingDataSetConfig"
        self.adaptor = adaptor
        self.version = version
        self.resolutions = resolutions
        self.playable = playable
        if playable:
            self.speeds = speeds
        self.exports = exports
        self.inputs = inputs
        self.metadata = metadata if metadata is not None else dict()
        self.filters = filters if filters is not None else dict()
        self.filterSemanticAnnotation = filterSemanticAnnotation if filterSemanticAnnotation is not None else dict()

    @classmethod
    def from_dict(cls, data):
        if data is None:
            return None
        if "@id" in data:
            del data["@id"]
        config = cls(None, None, None, None)
        for prop in cls.__annotations__.keys():
            attribute = data.get(prop)
            if attribute is not None:
                if prop in ['exports', 'inputs']:
                    attribute = [ImagingDataSetControl.from_dict(control) for control in attribute]
                elif prop in ['filters']:
                    filters = {}
                    for attr in attribute:
                        filters[attr] = [ImagingDataSetControl.from_dict(control) for control in attribute[attr]]
                    attribute = filters
                elif prop in ['filterSemanticAnnotation']:
                    filter_semantic_annotation = {}
                    for attr in attribute:
                        filter_semantic_annotation[attr] = ImagingSemanticAnnotation.from_dict(attribute[attr])
                    attribute = filter_semantic_annotation
            config.__dict__[prop] = attribute
        return config


class ImagingDataSetImage(AbstractImagingClass):
    config: ImagingDataSetConfig
    previews: list
    imageConfig: dict
    index: int
    metadata: dict

    def __init__(self, config: ImagingDataSetConfig, imageConfig=None, previews=None, metadata=None, index=0):
        self.__dict__["@type"] = "imaging.dto.ImagingDataSetImage"
        assert config is not None, "Config must not be None!"
        self.config = config
        self.imageConfig = imageConfig if imageConfig is not None else dict()
        self.previews = previews if previews is not None else [ImagingDataSetPreview("png")]
        self.metadata = metadata if metadata is not None else dict()
        self.index = index if index is not None else 0
        assert isinstance(self.previews, list), "Previews must be a list!"

    def add_preview(self, preview):
        self.previews += [preview]

    @classmethod
    def from_dict(cls, data):
        if data is None:
            return None
        if "@id" in data:
            del data["@id"]
        config = ImagingDataSetConfig.from_dict(data.get('config'))
        image = cls(config, None, None, None)
        for prop in cls.__annotations__.keys():
            attribute = data.get(prop)
            if prop == 'previews' and attribute is not None:
                attribute = [ImagingDataSetPreview.from_dict(preview) for preview in attribute]
            if prop not in ['config']:
                image.__dict__[prop] = attribute
        return image


class ImagingDataSetPropertyConfig(AbstractImagingClass):
    images: list
    metadata: dict

    def __init__(self, images: list, metadata=None):
        self.__dict__["@type"] = "imaging.dto.ImagingDataSetPropertyConfig"
        self.images = images if images is not None else []
        self.metadata = metadata if metadata is not None else dict()

    @classmethod
    def from_dict(cls, data: dict):
        assert data is not None and any(data), "There is no property config found!"
        if "@id" in data:
            del data["@id"]
        attr = data.get('images')
        images = [ImagingDataSetImage.from_dict(image) for image in attr] if attr is not None else None
        metadata = data.get('metadata')
        return cls(images, metadata)

    def add_image(self, image: ImagingDataSetImage):
        if self.images is None:
            self.images = []
        self.images += [image]